__all__ = ['DynamicDidDefinition']

from udsoncan.common.MemoryLocation import MemoryLocation
from udsoncan.common.AddressAndLengthFormatIdentifier import AddressAndLengthFormatIdentifier

from typing import Union, Type, List, Optional, cast


class DynamicDidDefinition:
    """
    This class serves as a container for the different pieces of a dynamic DID defined by the DynamicallyDefineDataIDentifier service.

    """

    class ByDidDefinition:
        source_did: int
        position: int
        memorysize: int

        def __init__(self, source_did: int, position: int, memorysize: int):
            if not isinstance(source_did, int):
                raise ValueError('source_did must be an integer')
            if source_did > 0xFFFF or source_did < 0:
                raise ValueError('source_did must be an integer between 0 and 0xFFFF')

            if not isinstance(position, int):
                raise ValueError('position must be an integer')
            if position < 0:
                raise ValueError('position must be an integer greater than 0')

            if not isinstance(memorysize, int):
                raise ValueError('memorysize must be an integer')
            if memorysize < 0:
                raise ValueError('memorysize must be an integer greater than 0')

            self.source_did = source_did
            self.position = position
            self.memorysize = memorysize

    class ByMemloc:
        memloc: MemoryLocation

        def __init__(self, memloc: MemoryLocation):
            if not isinstance(memloc, MemoryLocation):
                raise ValueError('memloc must be an instance of MemoryLocation')

            self.memloc = memloc

    entries: Union[List[ByDidDefinition], List[ByMemloc]]
    definition_type: Optional[Union[Type[ByDidDefinition], Type[ByMemloc]]]

    def __init__(self, *args, **kwargs):
        self.entries = list()
        self.definition_type = None

        if len(args) > 0 or len(kwargs) > 0:
            self.add(*args, **kwargs)

    def add(self, *args, **kwargs) -> None:
        """
        Add a piece of definition for a dynamic DID. 

        When defining by memory address, only ``memloc`` can be supplied. 
        Example : ``myDidDefinition.add(myMemLoc)``

        When defining with a source DID, these parameters must be supplied : ``source_did``, ``position``, ``memorysize``. 
        Example : ``myDidDefinition.add(source_did=1234, position=1, memorysize=2)``

        :param source_did: The source DID from which to fetch data from
        :type source_did: int

        :param position: Start position of the data to fetch inside the source DID data
        :type position: int

        :param memorysize: Length of data to fetch inside the source DID data
        :type memorysize: int

        :param memloc: MemoryLocation containing an address, a size and an encoding format
        :type memloc: :ref:`MemoryLocation<MemoryLocation>`

        """
        entry: Optional[Union["DynamicDidDefinition.ByMemloc", "DynamicDidDefinition.ByDidDefinition"]] = None

        if len(args) > 0:
            if isinstance(args[0], MemoryLocation):
                entry = self.ByMemloc(memloc=args[0])
        elif 'memloc' in kwargs:
            entry = self.ByMemloc(memloc=kwargs['memloc'])

        if entry is None:
            entry = self.ByDidDefinition(*args, **kwargs)

        if self.definition_type is None:
            self.definition_type = entry.__class__
        else:
            if entry.__class__ != self.definition_type:
                raise ValueError('It is not possible to define a composite dynamic DID by memory address and by source DID at the same time.')

        self.entries.append(entry)  # type: ignore

    def is_by_source_did(self) -> bool:
        return True if self.definition_type == self.ByDidDefinition else False

    def is_by_memory_address(self) -> bool:
        return True if self.definition_type == self.ByMemloc else False

    def get_alfid(self) -> AddressAndLengthFormatIdentifier:
        if not self.is_by_memory_address():
            raise ValueError('No AddressAndLengthFormatIdentifier available in this DynamicDidDefinition. Only available when defined by memory address')

        alfid: Optional[AddressAndLengthFormatIdentifier] = None
        for entry in cast(List["DynamicDidDefinition.ByMemloc"], self.entries):
            if alfid is None:
                alfid = entry.memloc.alfid
            else:
                if alfid.get_byte() != entry.memloc.alfid.get_byte():
                    raise ValueError(
                        'AddressAndLengthFormatIdentifier of DynamicDidDefinition entries are not consistent. Make sure address_format and memorysize_format are the same accross all entries.')

        if alfid is None:
            raise ValueError('No AddressAndLengthFormatIdentifier avaialble')  # should never happen.

        return alfid

    def get(self) -> Union[List[ByDidDefinition], List[ByMemloc]]:
        return self.entries



__all__ = ['Routine']

from typing import Optional


class Routine:
    """
    Defines a list of constants that are routine identifiers defined by the UDS standard.
    This class provides no functionality apart from defining these constants
    """
    DeployLoopRoutineID = 0xE200
    EraseMemory = 0xFF00
    CheckProgrammingDependencies = 0xFF01
    EraseMirrorMemoryDTCs = 0xFF02

    @classmethod
    def name_from_id(cls, routine_id: int) -> Optional[str]:
        # Helper to print the type of requests (logging purpose) as defined by ISO-14229:2006, Annex F
        if not isinstance(routine_id, int) or routine_id < 0 or routine_id > 0xFFFF:
            raise ValueError('Routine ID must be a valid integer between 0 and 0xFFFF')

        if routine_id >= 0x0000 and routine_id <= 0x00FF:
            return 'ISOSAEReserved'
        if routine_id >= 0x0100 and routine_id <= 0x01FF:
            return 'TachographTestIds'
        if routine_id >= 0x0200 and routine_id <= 0xDFFF:
            return 'VehicleManufacturerSpecific'
        if routine_id >= 0xE000 and routine_id <= 0xE1FF:
            return 'OBDTestIds'
        if routine_id == 0xE200:
            return 'DeployLoopRoutineID'
        if routine_id >= 0xE201 and routine_id <= 0xE2FF:
            return 'SafetySystemRoutineIDs'
        if routine_id >= 0xE300 and routine_id <= 0xEFFF:
            return 'ISOSAEReserved'
        if routine_id >= 0xF000 and routine_id <= 0xFEFF:
            return 'SystemSupplierSpecific'
        if routine_id == 0xFF00:
            return 'EraseMemory'
        if routine_id == 0xFF01:
            return 'CheckProgrammingDependencies'
        if routine_id == 0xFF02:
            return 'EraseMirrorMemoryDTCs'
        if routine_id >= 0xFF03 and routine_id <= 0xFFFF:
            return 'ISOSAEReserved'
        return None



from udsoncan import MemoryLocation
from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseResponseData
from udsoncan.ResponseCode import ResponseCode

from typing import cast


class ReadMemoryByAddress(BaseService):
    _sid = 0x23
    _use_subfunction = False

    supported_negative_response = [ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestOutOfRange,
                                   ResponseCode.SecurityAccessDenied
                                   ]

    class ResponseData(BaseResponseData):
        """
        .. data:: memory_block

                bytes object reflecting the content of the read memory
        """

        memory_block: bytes

        def __init__(self, memory_block: bytes):
            super().__init__(ReadMemoryByAddress)
            self.memory_block = memory_block

    class InterpretedResponse(Response):
        service_data: "ReadMemoryByAddress.ResponseData"

    @classmethod
    def make_request(cls, memory_location: MemoryLocation) -> Request:
        """
        Generates a request for ReadMemoryByAddress

        :param memory_location: The address and the size of the memory block to read.
        :type memory_location: :ref:`MemoryLocation <MemoryLocation>`

        :raises ValueError: If parameters are out of range, missing or wrong type
        """

        if not isinstance(memory_location, MemoryLocation):
            raise ValueError('Given memory location must be an instance of MemoryLocation')

        request = Request(service=cls)
        request.data = bytes()
        request.data += memory_location.alfid.get_byte()  # AddressAndLengthFormatIdentifier
        request.data += memory_location.get_address_bytes()
        request.data += memory_location.get_memorysize_bytes()

        return request

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`ReadMemoryByAddress.ResponseData<udsoncan.services.ReadMemoryByAddress.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :raises InvalidResponseException: If length of ``response.data`` is too short
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 1:
            raise InvalidResponseException(response, "Response data must be at least 1 byte")

        response.service_data = cls.ResponseData(
            memory_block=response.data
        )

        return cast(ReadMemoryByAddress.InterpretedResponse, response)



from udsoncan import DataIdentifier, Routine, Units
from test.UdsTest import UdsTest

class TestDefinitions(UdsTest):
    def test_data_identifier_name_from_id(self):
        for i in range(0x10000):
            name = DataIdentifier.name_from_id(i)
            self.assertTrue(isinstance(name, str))

    def test_routine_name_from_id(self):
        for i in range(0x10000):
            name = Routine.name_from_id(i)
            self.assertTrue(isinstance(name, str))




from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseSubfunction, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
import udsoncan.tools as tools

from typing import Optional, cast


class ControlDTCSetting(BaseService):
    _sid = 0x85

    supported_negative_response = [ResponseCode.SubFunctionNotSupported,
                                   ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestOutOfRange
                                   ]

    class SettingType(BaseSubfunction):
        """
        ControlDTCSetting defined subfunctions
        """

        __pretty_name__ = 'setting type'

        on = 1
        off = 2
        vehicleManufacturerSpecific = (0x40, 0x5F)  # To be able to print textual name for logging only.
        systemSupplierSpecific = (0x60, 0x7E)		# To be able to print textual name for logging only.

    class ResponseData(BaseResponseData):
        """
        .. data:: setting_type_echo

                Request subfunction echoed back by the server
        """

        def __init__(self, setting_type_echo: int):
            super().__init__(ControlDTCSetting)
            self.setting_type_echo = setting_type_echo

    class InterpretedResponse(Response):
        service_data: "ControlDTCSetting.ResponseData"

    @classmethod
    def make_request(cls, setting_type: int, data: Optional[bytes] = None) -> Request:
        """
        Generates a request for ControlDTCSetting

        :param setting_type: Service subfunction. Allowed values are from 0 to 0x7F
        :type setting_type: int

        :param data: Optional additional data sent with the request called `DTCSettingControlOptionRecord`
        :type data: bytes

        :raises ValueError: If parameters are out of range, missing or wrong type
        """

        tools.validate_int(setting_type, min=0, max=0x7F, name='Setting type')
        if data is not None:
            if not isinstance(data, bytes):
                raise ValueError('data must be a valid bytes object')

        return Request(service=cls, subfunction=setting_type, data=data)

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`ControlDTCSetting.ResponseData<udsoncan.services.ControlDTCSetting.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :raises InvalidResponseException: If length of ``response.data`` is too short
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 1:
            raise InvalidResponseException(response, "Response data must be at least 1 byte")

        response.service_data = cls.ResponseData(
            setting_type_echo=response.data[0]
        )

        return cast(ControlDTCSetting.InterpretedResponse, response)



__all__ = ['Filesize']
import math

from typing import Optional


class Filesize:
    """
    This class represent a file size used by the RequestFileTransfer service.

    :param uncompressed: Represent the uncompressed size in bytes
    :type uncompressed: int

    :param compressed: Represent the compressed size in bytes
    :type compressed: int

    :param width: The number of byte that should be used to encode the filesize in a payload
    :type width: int
    """

    uncompressed: Optional[int]
    compressed: Optional[int]
    width: int

    def __init__(self, uncompressed: Optional[int] = None, compressed: Optional[int] = None, width: Optional[int] = None):

        if uncompressed is None and compressed is None:
            raise ValueError('At least one size must be specified')

        if uncompressed is not None:
            if not isinstance(uncompressed, int):
                raise ValueError('Uncompressed size must be an integer')

            if uncompressed < 0:
                raise ValueError("Uncompressed size must be an integer greater than 0")

        if compressed is not None:
            if not isinstance(compressed, int):
                raise ValueError('Compressed size must be an integer')

            if compressed < 0:
                raise ValueError("Comrpessed size must be an integer greater than 0")

        if width is not None:
            if not isinstance(width, int):
                raise ValueError('Width must be an integer')

            if width < 0:
                raise ValueError("Width size must be an integer greater than 0")
            maxsize = 2**(width * 8) - 1

            if compressed is not None:
                if compressed > maxsize:
                    raise ValueError("With width=%d, compressed size must be smaller than %d" % (width, maxsize))

            if uncompressed is not None:
                if uncompressed > maxsize:
                    raise ValueError("With width=%d, uncompressed size must be smaller than %d" % (width, maxsize))
        else:
            width_value = 0
            if uncompressed is not None:
                width_value = max(width_value, uncompressed)

            if compressed is not None:
                width_value = max(width_value, compressed)
            width = math.ceil(math.log2(width_value + 1) / 8)

        self.uncompressed = uncompressed
        self.compressed = compressed
        self.width = width

    def get_width(self) -> int:
        return self.width

    def get_uncompressed_bytes(self) -> bytes:
        if self.uncompressed is not None:
            return self.uncompressed.to_bytes(self.width, byteorder='big')
        else:
            return b''

    def get_compressed_bytes(self) -> bytes:
        if self.compressed is not None:
            return self.compressed.to_bytes(self.width, byteorder='big')
        else:
            return b''

    def __str__(self) -> str:
        uncompressed_str = 'None' if self.uncompressed is None else '0x%02x' % self.uncompressed
        compressed_str = 'None' if self.compressed is None else '0x%02x' % self.compressed
        width_str = 'None' if self.width is None else '0x%02x' % self.width

        return "Filesize<Uncompressed=%s, Compressed=%s. Width=%s>" % (uncompressed_str, compressed_str, width_str)

    def __repr__(self) -> str:
        uncompressed_str = 'None' if self.uncompressed is None else '0x%02x' % self.uncompressed
        compressed_str = 'None' if self.compressed is None else '0x%02x' % self.compressed
        width_str = 'None' if self.width is None else '0x%02x' % self.width

        return "<Filesize: Uncompressed=%s, Compressed=%s. Width=%s at 0x%08x>" % (uncompressed_str, compressed_str, width_str, id(self))



import struct

from udsoncan import DidCodec, check_did_config, make_did_codec_from_definition, fetch_codec_definition_from_config, DIDConfig
from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
import udsoncan.tools as tools

from typing import Dict, Any, Union, List, cast, Optional


class ReadDataByIdentifier(BaseService):
    _sid = 0x22
    _use_subfunction = False

    supported_negative_response = [ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestOutOfRange,
                                   ResponseCode.SecurityAccessDenied
                                   ]

    class ResponseData(BaseResponseData):
        """
        .. data:: values

                Dictionary mapping the DID (int) with the value returned by the associated :ref:`DidCodec<DidCodec>`.decode method
        """

        values: Dict[int, Any]

        def __init__(self, values: Dict[int, Any]):
            super().__init__(ReadDataByIdentifier)

            self.values = values

    class InterpretedResponse(Response):
        service_data: "ReadDataByIdentifier.ResponseData"

    @classmethod
    def validate_didlist_input(cls, dids: Union[int, List[int]]) -> List[int]:
        if not isinstance(dids, int) and not isinstance(dids, list):
            raise ValueError("Data Identifier must either be an integer or a list of integer")

        if isinstance(dids, int):
            tools.validate_int(dids, min=0, max=0xFFFF, name='Data Identifier')

        if isinstance(dids, list):
            for did in dids:
                tools.validate_int(did, min=0, max=0xFFFF, name='Data Identifier')

        return [dids] if not isinstance(dids, list) else dids

    @classmethod
    def make_request(cls, didlist: Union[int, List[int]], didconfig: Optional[DIDConfig]) -> Request:
        """
        Generates a request for ReadDataByIdentifier

        :param didlist: List of data identifier to read.
        :type didlist: list[int]

        :param didconfig: Optional definition of DID codecs for validation. Dictionary mapping a DID (int) to a valid :ref:`DidCodec<DidCodec>` class or pack/unpack string 
        :type didconfig: dict[int] = :ref:`DidCodec<DidCodec>`

        :raises ValueError: If parameters are out of range, missing or wrong type
        :raises ConfigError: If didlist contains a DID not defined in didconfig
        """

        didlist = cls.validate_didlist_input(didlist)

        req = Request(cls)
        if didconfig is not None:
            # Return a validated did config. Format may change, entries might be added if default value is set.
            check_did_config(didlist, didconfig)

            did_reading_all_data = None
            for did in didlist:
                # Make sure the config is good before sending the request. This method can raise.
                codec_definition = fetch_codec_definition_from_config(did, didconfig)
                codec = make_did_codec_from_definition(codec_definition)

                try:
                    len(codec)  # Validate the length function. May raise
                    if did_reading_all_data is not None:
                        raise ValueError('Did 0x%04X is configured to read the rest of the payload (__len__ raisong ReadAllRemainingData), but a subsequent DID is requested (0x%04x)' % (
                            did_reading_all_data, did))
                except DidCodec.ReadAllRemainingData:
                    if did_reading_all_data is not None:
                        raise ValueError('It is impossible to read 2 DIDs configured to read the rest of the payload (__len__ raising ReadAllRemainingData). Dids are : 0x%04X and 0x%04X' % (
                            did_reading_all_data, did))
                    did_reading_all_data = did

        req.data = struct.pack('>' + 'H' * len(didlist), *didlist)  # Encode list of DID

        return req

    @classmethod
    def interpret_response(cls,
                           response: Response,
                           didlist: Union[int, List[int]],
                           didconfig: DIDConfig,
                           tolerate_zero_padding: bool = True) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`ReadDataByIdentifier.ResponseData<udsoncan.services.ReadDataByIdentifier.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :param didlist:  List of data identifiers used for the request.
        :type didlist: list[int]

        :param didconfig: Definition of DID codecs. Dictionary mapping a DID (int) to a valid :ref:`DidCodec<DidCodec>` class or pack/unpack string 
        :type didconfig: dict[int] = :ref:`DidCodec<DidCodec>`

        :param tolerate_zero_padding: Ignore trailing zeros in the response data avoiding raising false :class:`InvalidResponseException<udsoncan.exceptions.InvalidResponseException>`.
        :type tolerate_zero_padding: bool

        :raises ValueError: If parameters are out of range, missing or wrong type
        :raises ConfigError: If ``didlist`` parameter or response contains a DID not defined in ``didconfig``.
        :raises InvalidResponseException: If response data is incomplete or if DID data does not match codec length.
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        didlist = cls.validate_didlist_input(didlist)
        didconfig_validated = check_did_config(didlist, didconfig)

        response.service_data = cls.ResponseData(
            values={}
        )

        # Parsing algorithm to extract DID value
        offset = 0
        while True:
            if len(response.data) <= offset:
                break  # Done

            if len(response.data) <= offset + 1:
                if tolerate_zero_padding and response.data[-1] == 0:  # One extra byte, but it's a 0 and we accept that. So we're done
                    break
                raise InvalidResponseException(response, "Response given by server is incomplete.")

            did = struct.unpack('>H', response.data[offset:offset + 2])[0]  # Get the DID number
            if did == 0 and did not in didconfig_validated and tolerate_zero_padding:  # We read two zeros and that is not a DID bu we accept that. So we're done.
                if response.data[offset:] == b'\x00' * (len(response.data) - offset):
                    break

            codec_definition = fetch_codec_definition_from_config(did, didconfig_validated)
            codec = make_did_codec_from_definition(codec_definition)
            offset += 2

            try:
                payload_size = len(codec)
            except DidCodec.ReadAllRemainingData:
                payload_size = len(response.data) - offset

            if len(response.data) < offset + payload_size:
                raise InvalidResponseException(
                    response, "Value for data identifier 0x%04x was incomplete according to definition in configuration" % did)

            subpayload = response.data[offset:offset + payload_size]
            offset += payload_size  # Codec must define a __len__ function that matches the encoded payload length.
            val = codec.decode(subpayload)
            response.service_data.values[did] = val

        return cast(ReadDataByIdentifier.InterpretedResponse, response)



from udsoncan import Request
from udsoncan.BaseService import BaseService
from test.UdsTest import UdsTest

class DummyServiceNormal(BaseService):
    _sid = 0x13

class DummyServiceNoSubunction(BaseService):
    _sid = 0x13
    _use_subfunction = False


class TestRequest(UdsTest):

    def test_create_from_instance_ok(self):
        req = Request(DummyServiceNormal())
        self.assertEqual(req.service.request_id(), 0x13)

    def test_create_from_class_ok(self):
        req = Request(DummyServiceNormal, subfunction=0x44)
        self.assertEqual(req.service.request_id(), 0x13)

    def test_make_payload_basic(self):
        req = Request(DummyServiceNormal, subfunction=0x44)
        payload = req.get_payload()
        self.assertEqual(b"\x13\x44", payload)

    def test_make_payload_custom_data(self):
        req = Request(DummyServiceNormal, subfunction=0x44)
        req.data = b"\x12\x34\x56\x78"
        payload = req.get_payload()
        self.assertEqual(b"\x13\x44\x12\x34\x56\x78", payload)

    def test_make_payload_custom_data_no_subfunction(self):
        req = Request(DummyServiceNoSubunction, subfunction=0x44)
        req.data = b"\x12\x34\x56\x78"
        payload = req.get_payload()
        self.assertEqual(b"\x13\x12\x34\x56\x78", payload)			

    def test_suppress_positive_response(self):
        req = Request(DummyServiceNormal, subfunction=0x44, suppress_positive_response=True)
        payload = req.get_payload()
        self.assertEqual(b"\x13\xC4", payload)	# Subfunction bit 7 is set

    def test_suppress_positive_response_override(self):
        req = Request(DummyServiceNormal, subfunction=0x44, suppress_positive_response=False)
        payload = req.get_payload(suppress_positive_response=True)
        self.assertEqual(b"\x13\xC4", payload)	# Subfunction bit 7 is set	

        req = Request(DummyServiceNormal, subfunction=0x44, suppress_positive_response=True)
        payload = req.get_payload(suppress_positive_response=False)
        self.assertEqual(b"\x13\x44", payload)	# Subfunction bit 7 is cleared		

    def test_from_payload_basic(self):
        payload=b'\x3E\x01'	# 0x3e = TesterPresent
        req = Request.from_payload(payload)
        self.assertEqual(req.service.request_id(), 0x3E)
        self.assertEqual(req.subfunction, 0x01)
        self.assertFalse(req.suppress_positive_response, 0x01)

    def test_from_payload_suppress_positive_response(self):
        payload=b'\x3E\x81'	# 0x3e = TesterPresent
        req = Request.from_payload(payload)
        self.assertEqual(req.service.request_id(), 0x3E)
        self.assertEqual(req.subfunction, 0x01)
        self.assertTrue(req.suppress_positive_response)

    def test_from_payload_custom_data(self):
        payload=b'\x3E\x01\x12\x34\x56\x78'	# 0x3E = TesterPresent
        req = Request.from_payload(payload)
        self.assertEqual(req.service.request_id(), 0x3E)
        self.assertEqual(req.subfunction, 0x01)
        self.assertEqual(req.data, b'\x12\x34\x56\x78')

    def test_from_empty_payload(self):
        payload = b''
        req = Request.from_payload(payload)
        self.assertIsNone(req.service)
        self.assertIsNone(req.subfunction)
        self.assertIsNone(req.data)

    def test_from_bad_payload(self):
        payload = b'\xFF\xFF'
        req = Request.from_payload(payload)
        self.assertIsNone(req.service)
        self.assertIsNone(req.subfunction)
        self.assertIsNone(req.data)

    def test_str_repr(self):
        req = Request(DummyServiceNormal)
        str(req)
        req.__repr__()

    def test_from_input_param(self):
        with self.assertRaises(ValueError):
            req = Request("a string")	

        with self.assertRaises(ValueError):
            req = Request(DummyServiceNormal, "string")	

        with self.assertRaises(ValueError):
            req = Request(DummyServiceNormal(), data=123)	

    def test_spr_with_no_subfunction(self):
        with self.assertRaises(ValueError):
            Request(service=DummyServiceNoSubunction, suppress_positive_response=True)




import socket
import queue
import threading
import logging
import binascii
import sys
from abc import ABC, abstractmethod
import time
from typing import Union, Dict
import ctypes
import selectors

try:
    import can  # type:ignore
    _import_can_err = None
except Exception as e:
    _import_can_err = e

try:
    import isotp    # type:ignore
    _import_isotp_err = None
except Exception as e:
    _import_isotp_err = e

try:
    from udsoncan.j2534 import J2534, TxStatusFlag, Protocol_ID, Error_ID, Ioctl_Flags, Ioctl_ID, SCONFIG_LIST
    _import_j2534_err = None
except Exception as e:
    _import_j2534_err = e

try:
    from aioisotp.sync import SyncISOTPNetwork, SyncConnection  # type:ignore
    _import_aioisotp_err = None
except Exception as e:
    _import_aioisotp_err = e

from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan.exceptions import TimeoutException


from typing import Optional, Tuple, cast


class BaseConnection(ABC):

    name: str
    logger: logging.Logger

    def __init__(self, name: Optional[str] = None):
        if name is None:
            self.name = 'Connection'
        else:
            self.name = 'Connection[%s]' % (name)

        self.logger = logging.getLogger(self.name)

    def send(self, data: Union[bytes, Request, Response], timeout: Optional[float] = None) -> None:
        """Sends data to the underlying transport protocol

        :param data: The data or object to send. If a Request or Response is given, the value returned by get_payload() will be sent.
        :type data: bytes, Request, Response

        :returns: None
        """
        if not self.is_open():
            raise RuntimeError("Connection is not opened")

        if isinstance(data, Request) or isinstance(data, Response):
            payload = data.get_payload()
        else:
            payload = data

        self.logger.debug('Sending %d bytes : [%s]' % (len(payload), binascii.hexlify(payload).decode('ascii')))

        # backward compatibility
        if 'timeout' in self.specific_send.__code__.co_varnames:
            self.specific_send(payload, timeout=timeout)
        else:
            self.specific_send(payload)

    def wait_frame(self, timeout: Optional[float] = None, exception: bool = False) -> Optional[bytes]:
        """Waits for the reception of a frame of data from the underlying transport protocol

        :param timeout: The maximum amount of time to wait before giving up in seconds
        :type timeout: float
        :param exception: Boolean value indicating if this function may return exceptions.
                When ``True``, all exceptions may be raised, including ``TimeoutException``
                When ``False``, all exceptions will be logged as ``DEBUG`` and ``None`` will be returned.
        :type exception: bool

        :returns: Received data
        :rtype: bytes or None
        """
        if not self.is_open():
            raise RuntimeError("Connection is not opened")

        try:
            frame = self.specific_wait_frame(timeout=timeout)
        except Exception as e:
            self.logger.debug('No data received: [%s] - %s ' % (e.__class__.__name__, str(e)))

            if exception == True:
                raise
            else:
                frame = None

        if frame is not None:
            self.logger.debug('Received %d bytes : [%s]' % (len(frame), binascii.hexlify(frame).decode('ascii')))
        return frame

    def __enter__(self):
        return self

    @abstractmethod
    def specific_send(self, payload: bytes, timeout: Optional[float] = None) -> None:
        """The implementation of the send method.

        :param payload: Data to send
        :type payload: bytes

        :returns: None
        """
        pass

    @abstractmethod
    def specific_wait_frame(self, timeout: Optional[float] = None) -> Optional[bytes]:
        """The implementation of the ``wait_frame`` method. 

        :param timeout: The maximum amount of time to wait before giving up
        :type timeout: float

        :returns: Received data
        :rtype: bytes or None
        """
        pass

    @abstractmethod
    def open(self) -> "BaseConnection":
        """ Set up the connection object. 

        :returns: None
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """ Close the connection object

        :returns: None
        """
        pass

    @abstractmethod
    def empty_rxqueue(self) -> None:
        """ Empty all unread data in the reception buffer.

        :returns: None
        """
        pass

    @abstractmethod
    def is_open(self) -> bool:
        """ Tells if the connection is open.

        :returns: bool
        """
        pass

    def __exit__(self, type, value, traceback):
        pass


class SocketConnection(BaseConnection):
    """
    Sends and receives data through a socket.

    :param sock: The socket to use. This socket must be bound and ready to use. Only ``send()`` and ``recv()`` will be called by this Connection
    :type sock: socket.socket
    :param bufsize: Maximum buffer size of the socket, this value is passed to ``recv()``
    :type bufsize: int
    :param name: This name is included in the logger name so that its output can be redirected. The logger name will be ``Connection[<name>]``
    :type name: string

    """

    rxqueue: "queue.Queue[bytes]"
    exit_requested: bool
    opened: bool
    rxthread: Optional[threading.Thread]
    sock: socket.socket
    bufsize: int

    def __init__(self, sock: socket.socket, bufsize: int = 4095, name: Optional[str] = None):
        BaseConnection.__init__(self, name)

        self.rxqueue = queue.Queue()
        self.exit_requested = False
        self.opened = False
        self.rxthread = None
        self.sock = sock
        self.bufsize = bufsize

    def open(self) -> "SocketConnection":
        self.exit_requested = False
        self.rxthread = threading.Thread(target=self.rxthread_task, daemon=True)
        self.rxthread.start()
        self.opened = True
        self.logger.info('Connection opened')
        return self

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def is_open(self) -> bool:
        return self.opened

    def rxthread_task(self) -> None:
        sel = selectors.DefaultSelector()
        sel.register(self.sock, selectors.EVENT_READ)
        while not self.exit_requested:
            try:
                events = sel.select(timeout=0.2)
                if events:
                    data = self.sock.recv(self.bufsize)
                    if data is not None:
                        self.rxqueue.put(data)
            except Exception:
                self.exit_requested = True

    def close(self) -> None:
        self.exit_requested = True
        if self.rxthread is not None:
            self.rxthread.join()
        self.opened = False
        self.logger.info('Connection closed')

    def specific_send(self, payload: bytes, timeout: Optional[float] = None) -> None:
        # timeout not used for generic sockets
        self.sock.send(payload)

    def specific_wait_frame(self, timeout: Optional[float] = None) -> Optional[bytes]:
        if not self.opened:
            raise RuntimeError("Connection is not open")

        timedout = False
        frame = None
        try:
            frame = self.rxqueue.get(block=True, timeout=timeout)
        except queue.Empty:
            timedout = True

        if timedout:
            raise TimeoutException("Did not received frame in time (timeout=%s sec)" % timeout)

        return frame

    def empty_rxqueue(self) -> None:
        while not self.rxqueue.empty():
            self.rxqueue.get()


class IsoTPSocketConnection(BaseConnection):
    """
    Sends and receives data through an ISO-TP socket. Makes cleaner code than SocketConnection but offers no additional functionality.
    The `can-isotp module <https://github.com/pylessard/python-can-isotp>`_ must be installed in order to use this connection

    :param interface: The can interface to use (example: ``can0``)
    :type interface: string
    :param address: The address used to bind the the socket. Before 1.21, txid/rxid were needed here, this has changed with v1.21
    :type address: ``isotp.Address`` or ``isotp.AsymmetricAddress`` 
    :param name: This name is included in the logger name so that its output can be redirected. The logger name will be ``Connection[<name>]``
    :type name: string
    :param tpsock: An optional ISO-TP socket to use instead of creating one.
    :type tpsock: isotp.socket

    """

    interface: str
    address: Union["isotp.Address", "isotp.AsymmetricAddress"]
    rxqueue: "queue.Queue[bytes]"
    exit_requested: bool
    opened: bool

    def __init__(self,
                 interface: str,
                 address: Union["isotp.Address", "isotp.AsymmetricAddress"],
                 name: Optional[str] = None,
                 tpsock: Optional["isotp.socket"] = None,
                 **kwargs
                 ):

        BaseConnection.__init__(self, name)

        self.interface = interface
        self.address = address
        self.rxqueue = queue.Queue()
        self.exit_requested = False
        self.opened = False

        # Lives with the past.
        if 'txid' in kwargs or 'rxid' in kwargs:
            raise RuntimeError(
                "Provide an isotp.Address to the %s. The interface has changed in a non-backward compatible way and this is now required." % self.__class__.__name__)

        if tpsock is None:
            if 'isotp' not in sys.modules:
                if _import_isotp_err is None:
                    raise ImportError('isotp module is not loaded')
                else:
                    raise _import_isotp_err
            self.tpsock = isotp.socket(timeout=0.1)
        else:
            self.tpsock = tpsock

    def open(self) -> "IsoTPSocketConnection":
        self.tpsock.bind(self.interface, address=self.address)
        self.exit_requested = False
        self.rxthread = threading.Thread(target=self.rxthread_task, daemon=True)
        self.rxthread.start()
        self.opened = True
        self.logger.info('Connection opened')
        return self

    def __enter__(self) -> "IsoTPSocketConnection":
        return self

    def __exit__(self, type, value, traceback) -> None:
        self.close()

    def is_open(self) -> bool:
        return self.tpsock.bound

    def rxthread_task(self) -> None:
        sel = selectors.DefaultSelector()
        sel.register(self.tpsock._socket, selectors.EVENT_READ)
        while not self.exit_requested:
            try:
                events = sel.select(timeout=0.2)
                if events:
                    data = self.tpsock.recv()
                    if data is not None:
                        self.rxqueue.put(data)
            except Exception:
                self.exit_requested = True

    def close(self) -> None:
        self.exit_requested = True
        if self.rxthread is not None:
            self.rxthread.join()
        self.tpsock.close()
        self.opened = False
        self.logger.info('Connection closed')

    def specific_send(self, payload: bytes, timeout: Optional[float] = None) -> None:
        self.tpsock.send(payload)

    def specific_wait_frame(self, timeout: Optional[float] = None) -> Optional[bytes]:
        if not self.opened:
            raise RuntimeError("Connection is not open")

        timedout = False
        frame = None
        try:
            frame = self.rxqueue.get(block=True, timeout=timeout)

        except queue.Empty:
            timedout = True

        if timedout:
            raise TimeoutException("Did not received ISOTP frame in time (timeout=%s sec)" % timeout)

        return frame

    def empty_rxqueue(self) -> None:
        while not self.rxqueue.empty():
            self.rxqueue.get()


class IsoTPConnection(IsoTPSocketConnection):
    """
    Same as :class:`IsoTPSocketConnection <udsoncan.connections.IsoTPSocketConnection.Session>`. Exists only for backward compatibility. 
    """
    pass


class QueueConnection(BaseConnection):
    """
    Sends and receives data using 2 Python native queues.

    - ``MyConnection.fromuserqueue`` : Data read from this queue when ``wait_frame`` is called
    - ``MyConnection.touserqueue`` : Data written to this queue when ``send`` is called

    :param mtu: Optional maximum frame size. Messages will be truncated to this size
    :type mtu: int
    :param name: This name is included in the logger name so that its output can be redirected. The logger name will be ``Connection[<name>]``
    :type name: string
    """

    fromuserqueue: "queue.Queue[bytes]"
    touserqueue: "queue.Queue[bytes]"
    opened: bool
    mtu: int

    def __init__(self, name: Optional[str] = None, mtu: int = 4095):
        BaseConnection.__init__(self, name)

        self.fromuserqueue = queue.Queue()  # Client reads from this queue. Other end is simulated
        self.touserqueue = queue.Queue()  # Client writes to this queue. Other end is simulated
        self.opened = False
        self.mtu = mtu

    def open(self) -> "QueueConnection":
        self.opened = True
        self.logger.info('Connection opened')
        return self

    def __enter__(self) -> "QueueConnection":
        return self

    def __exit__(self, type, value, traceback) -> None:
        self.close()

    def is_open(self) -> bool:
        return self.opened

    def close(self) -> None:
        self.empty_rxqueue()
        self.empty_txqueue()
        self.opened = False
        self.logger.info('Connection closed')

    def specific_send(self, payload: bytes, timeout: Optional[float] = None) -> None:
        if self.mtu is not None:
            if len(payload) > self.mtu:
                self.logger.warning("Truncating payload to be set to a length of %d" % (self.mtu))
                payload = payload[0:self.mtu]

        self.touserqueue.put(payload, block=True, timeout=timeout)

    def specific_wait_frame(self, timeout: Optional[float] = None) -> Optional[bytes]:
        if not self.opened:
            raise RuntimeError("Connection is not open")

        timedout = False
        frame = None
        try:
            frame = self.fromuserqueue.get(block=True, timeout=timeout)
        except queue.Empty:
            timedout = True

        if timedout:
            raise TimeoutException("Did not receive frame from user queue in time (timeout=%s sec)" % timeout)

        if self.mtu is not None:
            if frame is not None and len(frame) > self.mtu:
                self.logger.warning("Truncating received payload to a length of %d" % (self.mtu))
                frame = frame[0:self.mtu]

        return frame

    def empty_rxqueue(self) -> None:
        while not self.fromuserqueue.empty():
            self.fromuserqueue.get()

    def empty_txqueue(self) -> None:
        while not self.touserqueue.empty():
            self.touserqueue.get()


class PythonIsoTpConnection(BaseConnection):
    """
    Sends and receives data using a `can-isotp <https://github.com/pylessard/python-can-isotp>`_ Python module which is a Python implementation of the IsoTp transport protocol
    which can be coupled with `python-can <https://python-can.readthedocs.io>`_ module to interract with CAN hardware

    `can-isotp <https://github.com/pylessard/python-can-isotp>`_ must be installed in order to use this connection.

    See an :ref:`example<example_using_python_can>`

    :param isotp_layer: The IsoTP Transport layer object coming from the ``isotp`` module.
    :type isotp_layer: :class:`isotp.TransportLayer<isotp.TransportLayer>`

    :param name: This name is included in the logger name so that its output can be redirected. The logger name will be ``Connection[<name>]``
    :type name: string

    """

    subconn: Union["PythonIsoTpV1Connection", "PythonIsoTpV2Connection"]

    def __init__(self,
                 isotp_layer: Union["isotp.TransportLayerLogic", "isotp.TransportLayer"],
                 name: Optional[str] = None
                 ):
        BaseConnection.__init__(self, name)
        import isotp
        if hasattr(isotp, '_major_version_'):    # isotp v2.x
            if isotp._major_version_ == 2:
                if isinstance(isotp_layer, isotp.TransportLayer):   # This one has its own thread
                    self.subconn = PythonIsoTpV2Connection(isotp_layer, name)
                elif isinstance(isotp_layer, isotp.TransportLayerLogic):    # Need to create a thread for this one
                    self.subconn = PythonIsoTpV1Connection(isotp_layer, name)
                else:
                    raise ValueError("Invalid isotp layer object")
            else:
                raise NotImplementedError("Unsupported isotp version")
        else:   # isotp v1.x
            self.subconn = PythonIsoTpV1Connection(isotp_layer, name)

    def open(self) -> "PythonIsoTpConnection":
        self.subconn.open()
        return self

    def __enter__(self) -> "PythonIsoTpConnection":
        self.subconn.__enter__()
        return self

    def __exit__(self, type, value, traceback) -> None:
        return self.subconn.__exit__(type, value, traceback)

    def is_open(self) -> bool:
        return self.subconn.is_open()

    def close(self) -> None:
        return self.subconn.close()

    def specific_send(self, payload: bytes, timeout: Optional[float] = None) -> None:
        self.subconn.specific_send(payload, timeout)

    def specific_wait_frame(self, timeout: Optional[float] = None) -> Optional[bytes]:
        return self.subconn.specific_wait_frame(timeout)

    def empty_rxqueue(self) -> None:
        return self.subconn.empty_rxqueue()

    def empty_txqueue(self) -> None:
        return self.subconn.empty_txqueue()


class PythonIsoTpV2Connection(BaseConnection):

    isotp_layer: "isotp.TransportLayer"
    opened: bool

    def __init__(self, isotp_layer: "isotp.TransportLayer", name: Optional[str] = None):
        BaseConnection.__init__(self, name)
        self.opened = False
        self.isotp_layer = isotp_layer

        assert isinstance(self.isotp_layer, isotp.TransportLayer), 'isotp_layer must be a valid isotp.TransportLayer '

    def open(self) -> "PythonIsoTpV2Connection":
        self.isotp_layer.start()
        self.opened = True
        self.logger.info('Connection opened')
        return self

    def __enter__(self) -> "PythonIsoTpV2Connection":
        return self

    def __exit__(self, type, value, traceback) -> None:
        self.close()

    def is_open(self) -> bool:
        return self.opened

    def close(self) -> None:
        self.isotp_layer.stop()
        self.empty_rxqueue()
        self.empty_txqueue()
        self.opened = False
        self.logger.info('Connection closed')

    def specific_send(self, payload: bytes, timeout: Optional[float] = None) -> None:
        self.isotp_layer.send(payload, send_timeout=timeout)

    def specific_wait_frame(self, timeout: Optional[float] = None) -> Optional[bytes]:
        if not self.opened:
            raise RuntimeError("Connection is not opened")

        frame = self.isotp_layer.recv(block=True, timeout=timeout)
        if frame is None:
            raise TimeoutException("Did not receive IsoTP frame from the Transport layer in time (timeout=%s sec)" % timeout)

        return bytes(frame)

    def empty_rxqueue(self) -> None:
        self.isotp_layer.stop_receiving()
        self.isotp_layer.clear_rx_queue()

    def empty_txqueue(self) -> None:
        self.isotp_layer.stop_sending()
        self.isotp_layer.clear_tx_queue()


class PythonIsoTpV1Connection(BaseConnection):
    toIsoTPQueue: "queue.Queue[bytes]"
    fromIsoTPQueue: "queue.Queue[bytes]"
    rxthread: Optional[threading.Thread]
    exit_requested: bool
    opened: bool
    isotp_layer: "isotp.TransportLayerLogic"

    def __init__(self, isotp_layer: "isotp.TransportLayerLogic", name: Optional[str] = None):
        BaseConnection.__init__(self, name)
        self.toIsoTPQueue = queue.Queue()
        self.fromIsoTPQueue = queue.Queue()
        self.rxthread = None
        self.exit_requested = False
        self.opened = False
        self.isotp_layer = isotp_layer

        # isotp v1 TransportLayer == isotpv2.TransportLayerLogic
        if hasattr(isotp, 'TransportLayerLogic'):
            assert isinstance(self.isotp_layer, isotp.TransportLayerLogic), 'isotp_layer must be a valid isotp.TransportLayerLogic'
        else:
            assert isinstance(self.isotp_layer, isotp.TransportLayer), 'isotp_layer must be a valid isotp.isotp.TransportLayer'

    def open(self) -> "PythonIsoTpV1Connection":
        self.exit_requested = False
        self.rxthread = threading.Thread(target=self.rxthread_task, daemon=True)
        self.rxthread.start()
        self.opened = True
        self.logger.info('Connection opened')
        return self

    def __enter__(self) -> "PythonIsoTpV1Connection":
        return self

    def __exit__(self, type, value, traceback) -> None:
        self.close()

    def is_open(self) -> bool:
        return self.opened

    def close(self) -> None:
        self.empty_rxqueue()
        self.empty_txqueue()
        self.exit_requested = True
        if self.rxthread is not None:
            self.rxthread.join()
        self.isotp_layer.reset()
        self.opened = False
        self.logger.info('Connection closed')

    def specific_send(self, payload: bytes, timeout: Optional[float] = None):
        self.toIsoTPQueue.put(bytearray(payload))  # isotp.protocol.TransportLayer uses byte array. udsoncan is strict on bytes format

    def specific_wait_frame(self, timeout: Optional[float] = None) -> Optional[bytes]:
        if not self.opened:
            raise RuntimeError("Connection is not open")

        timedout = False
        frame = None
        try:
            frame = self.fromIsoTPQueue.get(block=True, timeout=timeout)
        except queue.Empty:
            timedout = True

        if timedout:
            raise TimeoutException("Did not receive IsoTP frame from the Transport layer in time (timeout=%s sec)" % timeout)

        if frame is None:
            return None

        # isotp.protocol.TransportLayer uses bytearray. udsoncan is strict on bytes format
        return bytes(frame)

    def empty_rxqueue(self) -> None:
        while not self.fromIsoTPQueue.empty():
            self.fromIsoTPQueue.get()

    def empty_txqueue(self) -> None:
        while not self.toIsoTPQueue.empty():
            self.toIsoTPQueue.get()

    def rxthread_task(self) -> None:
        while not self.exit_requested:
            try:
                while not self.toIsoTPQueue.empty():
                    self.isotp_layer.send(self.toIsoTPQueue.get())

                self.isotp_layer.process()

                while self.isotp_layer.available():
                    self.fromIsoTPQueue.put(self.isotp_layer.recv())

                time.sleep(self.isotp_layer.sleep_time())

            except Exception as e:
                self.exit_requested = True
                self.logger.error(str(e))


class J2534Connection(BaseConnection):
    """
    Sends and receives data through a J2534 Interface. 
    A windows DLL and a J2534 interface must be installed in order to use this connection

    :param windll: The path to the windows DLL for the J2534 interface (example: 'C:/Program Files{x86}../../openport 2.0/op20pt32.dll')
    :type interface: string
    :param rxid: The reception CAN id
    :type rxid: int 
    :param txid: The transmission CAN id
    :type txid: int
    :param name: This name is included in the logger name so that its output can be redirected. The logger name will be ``Connection[<name>]``
    :type name: string
    :param debug: This will enable windows debugging mode in the dll (see tactrix doc for additional information)
    :type debug: boolean
    :param args: Optional parameters list (Unused right now).
    :type args: list
    :param kwargs: Optional parameters dictionary Unused right now).
    :type kwargs: dict

    """

    interface: "J2534"
    protocol: "Protocol_ID"
    baudrate: int
    result: "Error_ID"
    firmwareVersion: "ctypes.Array[ctypes.c_char]"
    dllVersion: "ctypes.Array[ctypes.c_char]"
    apiVersion: "ctypes.Array[ctypes.c_char]"
    rxqueue: "queue.Queue[bytes]"
    exit_requested: bool
    opened: bool

    def __init__(self, windll: str, rxid: int, txid: int, name: Optional[str] = None, debug: bool = False, *args, **kwargs):
        BaseConnection.__init__(self, name)

        # Set up a J2534 interface using the DLL provided
        try:
            self.interface = J2534(windll=windll, rxid=rxid, txid=txid)
        except FileNotFoundError:
            raise RuntimeError('DLL not found')

        # Set the protocol to ISO15765, Baud rate to 500000
        self.protocol = Protocol_ID.ISO15765
        self.baudrate = 500000
        self.debug = debug

        try:
            # Open the interface (connect to the DLL)
            self.result, self.devID = self.interface.PassThruOpen()
        except OSError as e:
            if e.errno in [0x16, 0xe06d7363]:
                raise RuntimeError('J2534 Device busy')
            exception_str = type(e).__name__
            if e.errno is not None:
                exception_str += ', %s' % e.errno
            raise RuntimeError(exception_str)

        self.log_last_operation("PassThruOpen", with_raise=True)

        if debug:
            self.result = self.interface.PassThruIoctl(0,
                                                       Ioctl_Flags.TX_IOCTL_SET_DLL_DEBUG_FLAGS,
                                                       SCONFIG_LIST([(0, Ioctl_Flags.TX_IOCTL_DLL_DEBUG_FLAG_J2534_CALLS.value)])
                                                       )
            self.log_last_operation("PassThruIoctl SET_DLL_DEBUG")

        # Get the firmeware and DLL version etc, mainly for debugging output
        self.result, self.firmwareVersion, self.dllVersion, self.apiVersion = self.interface.PassThruReadVersion(self.devID)
        self.logger.info("J2534 FirmwareVersion: " + str(self.firmwareVersion.value) + ", dllVersoin: " +
                         str(self.dllVersion.value) + ", apiVersion" + str(self.apiVersion.value))

        # get the channel ID of the interface (used for subsequent communication)
        self.result, self.channelID = self.interface.PassThruConnect(self.devID, self.protocol.value, self.baudrate)
        self.log_last_operation("PassThruConnect", with_raise=True)

        configs = SCONFIG_LIST([
            (Ioctl_ID.DATA_RATE.value, 500000),
            (Ioctl_ID.LOOPBACK.value, 0),
            (Ioctl_ID.ISO15765_BS.value, 0x20),
            (Ioctl_ID.ISO15765_STMIN.value, 0),
        ])
        self.result = self.interface.PassThruIoctl(self.channelID, Ioctl_ID.SET_CONFIG, configs)
        self.log_last_operation("PassThruIoctl SET_CONFIG")

        self.result = self.interface.PassThruIoctl(self.channelID, Ioctl_ID.CLEAR_MSG_FILTERS)
        self.log_last_operation("PassThruIoctl CLEAR_MSG_FILTERS")

        # Set the filters and clear the read buffer (filters will be set based on tx/rxids)
        self.result = self.interface.PassThruStartMsgFilter(self.channelID, self.protocol.value)
        self.log_last_operation("PassThruStartMsgFilter")

        self.result = self.interface.PassThruIoctl(self.channelID, Ioctl_ID.CLEAR_RX_BUFFER)
        self.log_last_operation("PassThruIoctl CLEAR_RX_BUFFER")

        self.result = self.interface.PassThruIoctl(self.channelID, Ioctl_ID.CLEAR_TX_BUFFER)
        self.log_last_operation("PassThruIoctl CLEAR_TX_BUFFER")

        self.rxqueue = queue.Queue()
        self.exit_requested = False
        self.opened = False

    def open(self) -> "J2534Connection":
        self.exit_requested = False
        self.sem = threading.Semaphore()
        self.rxthread = threading.Thread(target=self.rxthread_task, daemon=True)
        self.rxthread.start()
        self.opened = True
        self.logger.info('J2534 Connection opened')
        return self

    def __enter__(self) -> "J2534Connection":
        return self

    def __exit__(self, type, value, traceback) -> None:
        self.close()

    def is_open(self) -> bool:
        return self.opened

    def rxthread_task(self) -> None:
        while not self.exit_requested:
            self.sem.acquire()
            try:
                result, data, numMessages = self.interface.PassThruReadMsgs(self.channelID, self.protocol.value, 1, 1)
                if data is not None:
                    self.rxqueue.put(data)
            except Exception:
                self.logger.critical("Exiting J2534 rx thread")
                self.exit_requested = True
            self.sem.release()
            time.sleep(0.001)

    def log_last_operation(self, exec_method: str, with_raise = False) -> None:
        if self.result != Error_ID.ERR_SUCCESS:
            res, pErrDescr = self.interface.PassThruGetLastError()
            err = "J2534 %s: %s (%s)" % (exec_method, pErrDescr, self.result)
            self.logger.error(err)
            if with_raise:
                raise RuntimeError(err)
            return

        elif self.debug:
            self.logger.debug("J2534 %s: OK" % (exec_method))

    def close(self) -> None:
        self.opened = False
        self.exit_requested = True
        self.rxthread.join()

        self.result = self.interface.PassThruDisconnect(self.channelID)
        self.log_last_operation('PassThruDisconnect')

        self.interface.PassThruClose(self.devID)
        self.log_last_operation('PassThruClose')

    def specific_send(self, payload: bytes, timeout: Optional[float] = None):
        timeout = 0 if timeout is None else timeout

        # Fix for avoid ERR_CONCURRENT_API_CALL. Stop reading
        self.sem.acquire()
        self.result = self.interface.PassThruWriteMsgs(self.channelID, payload, self.protocol.value, Timeout=int(timeout * 1000))
        self.log_last_operation('PassThruWriteMsgs', with_raise=True)
        self.sem.release()

    def specific_wait_frame(self, timeout: Optional[float] = None) -> Optional[bytes]:
        if not self.opened:
            raise RuntimeError("J2534 Connection is not open")

        timedout = False
        frame = None
        try:
            frame = self.rxqueue.get(block=True, timeout=timeout)
        except queue.Empty:
            timedout = True

        if timedout:
            raise TimeoutException("Did not received response from J2534 RxQueue (timeout=%s sec)" % timeout)

        return frame

    def empty_rxqueue(self) -> None:
        while not self.rxqueue.empty():
            self.rxqueue.get()

    def read_vbatt(self, digits=1) -> float:
        vbatt = ctypes.POINTER(ctypes.c_int32)()

        self.result = self.interface.PassThruIoctl(self.channelID, Ioctl_ID.READ_VBATT, None, vbatt)
        self.log_last_operation("PassThruIoctl READ_VBATT")

        value = ctypes.cast(vbatt, ctypes.c_void_p).value

        return round(value / 1000, digits) if value else 0


class FakeConnection(BaseConnection):
    """
    Sends and receives static data defined in a local dict. 
    Used so that an application can be tested without a live can network
    """

    rxqueue: "queue.Queue[bytes]"
    exit_requested: bool
    opened: bool
    ResponseData: Dict[bytes, bytes]

    def __init__(self, name=None, debug=False, *args, **kwargs):

        BaseConnection.__init__(self, name)

        self.rxqueue = queue.Queue()

        self.exit_requested = False
        self.opened = False

        self.ResponseData = {b'\x10\x03': b'\x50\x03\x12\x23\x34\x45',
                             b'\x22\xf1\x90\xf1\x89\xf1\x91\xf8\x06\xf1\xa3': b'\x22\xf1\x90\xf1\x89\xf1\x91\xf8\x06\xf1\xa3'}

    def open(self) -> "FakeConnection":
        self.opened = True
        self.logger.info('Fake Connection opened')
        return self

    def __enter__(self) -> "FakeConnection":
        return self

    def __exit__(self, type, value, traceback) -> None:
        self.close()

    def is_open(self) -> bool:
        return self.opened

    def close(self) -> None:
        self.exit_requested = True
        self.opened = False
        self.logger.info('Fake Connection closed')

    def specific_send(self, payload: bytes, timeout: Optional[float] = None):
        self.rxqueue.put(self.ResponseData[payload])

    def specific_wait_frame(self, timeout: Optional[float] = None) -> Optional[bytes]:
        if not self.opened:
            raise RuntimeError("Fake Connection is not open")

        timedout = False
        frame = None
        try:
            frame = self.rxqueue.get(block=True, timeout=timeout)
        except queue.Empty:
            timedout = True

        if timedout:
            raise TimeoutException("Did not received response from J2534 RxQueue (timeout=%s sec)" % timeout)

        return frame

    def empty_rxqueue(self) -> None:
        while not self.rxqueue.empty():
            self.rxqueue.get()


class SyncAioIsotpConnection(BaseConnection):
    """
    A wrapper for aioisotp sync variant

    `aioisotp <https://github.com/christiansandberg/aioisotp>`_ must be installed in order to use this connection.

    See an :ref:`example<example_using_aioisotp>`

    :param rxid: The reception CAN id
    :type rxid: int

    :param txid: The transmission CAN id
    :type txid: int

    :param name: This name is included in the logger name so that its output can be redirected. The logger name will be ``Connection[<name>]``
    :type name: string

    :param args: Optional parameters list passed to aioisotp binding method.
    :type args: list

    :param kwargs: Optional parameters dictionary passed to aioisotp binding method.
    :type kwargs: dict
    """

    network: "SyncISOTPNetwork"
    opened: bool
    rx_id: int
    tx_id: int
    conn: Optional["SyncConnection"]

    def __init__(self, rx_id: int, tx_id: int, name: Optional[str] = None, *args, **kwargs):
        BaseConnection.__init__(self, name)
        self.network = SyncISOTPNetwork(*args, **kwargs)
        self.opened = False
        self.rx_id = rx_id
        self.tx_id = tx_id
        self.conn = None
        self.opened = False

    def specific_send(self, payload: bytes, timeout: Optional[float] = None) -> None:
        if self.conn is None or not self.opened:
            raise RuntimeError("Connection is not opened")
        self.conn.send(payload)

    def specific_wait_frame(self, timeout: Optional[float] = None) -> Optional[bytes]:
        if not self.opened or self.conn is None:
            raise RuntimeError("Connection is not open")

        frame = cast(Optional[bytes], self.conn.recv(timeout))

        if frame is None and timeout:
            raise TimeoutException("Did not received frame in time (timeout=%s sec)" % timeout)

        return frame

    def open(self) -> "SyncAioIsotpConnection":
        self.network.open()
        self.conn = self.network.create_sync_connection(self.rx_id, self.tx_id)
        self.opened = True
        self.logger.info("Connection opened")
        return self

    def close(self) -> None:
        self.network.close()
        self.opened = False
        self.logger.info("Connection closed")

    def empty_rxqueue(self) -> None:
        if self.conn is not None:
            self.conn.empty()

    def is_open(self) -> bool:
        return self.opened

    def __enter__(self) -> "SyncAioIsotpConnection":
        return self

    def __exit__(self, type, value, traceback) -> None:
        self.close()



import struct
import math
from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan import IOMasks, IOValues, DidCodec, CodecDefinition, IOConfig, IOConfigEntry
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseSubfunction, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
import udsoncan.tools as tools
from udsoncan.common.dids import make_did_codec_from_definition, fetch_codec_definition_from_config

from typing import Optional, Any, Union, List, Dict, cast


class InputOutputControlByIdentifier(BaseService):
    _sid = 0x2F
    _use_subfunction = False

    class ControlParam(BaseSubfunction):
        """
        InputOutputControlByIdentifier defined control parameters as defined by ISO-14229:2006, Annex E
        """

        __pretty_name__ = 'control parameter'

        returnControlToECU = 0
        resetToDefault = 1
        freezeCurrentState = 2
        shortTermAdjustment = 3

    supported_negative_response = [ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestOutOfRange,
                                   ResponseCode.SecurityAccessDenied
                                   ]

    class ResponseData(BaseResponseData):
        """
        .. data:: did_echo

                DID echoed back by the server

        .. data:: control_param_echo

                control_param echoed back by the server

        .. data:: decoded_data

                Value processed by the :ref:`DidCodec<DidCodec>`.decode() method
        """

        did_echo: int
        control_param_echo: Optional[int]
        decoded_data: Optional[Any]

        def __init__(self, did_echo: int, control_param_echo: Optional[int] = None, decoded_data: Optional[Any] = None):
            super().__init__(InputOutputControlByIdentifier)
            self.did_echo = did_echo
            self.control_param_echo = control_param_echo
            self.decoded_data = decoded_data

    class InterpretedResponse(Response):
        service_data: "InputOutputControlByIdentifier.ResponseData"

    @classmethod
    def make_request(cls,
                     did: int,
                     control_param: Optional[int] = None,
                     values: Optional[Union[List[Any], Dict[str, Any], IOValues]] = None,
                     masks: Optional[Union[List[str], Dict[str, bool], IOMasks, bool]] = None,
                     ioconfig: Optional[IOConfig] = None
                     ) -> Request:
        """
        Generates a request for InputOutputControlByIdentifier

        :param did: Data identifier to represent the IO
        :type did: int

        :param control_param: Optional parameter that can be a value defined in :class:`InputOutputControlByIdentifier.ControlParam<udsoncan.services.InputOutputControlByIdentifier.ControlParam>`
        :type control_param: int

        :param values: Optional values to send to the server. This parameter will be given to :ref:`DidCodec<DidCodec>`.encode() method. 
                It can be:

                        - A list for positional arguments
                        - A dict for named arguments
                        - An instance of :ref:`IOValues<IOValues>` for mixed arguments

        :type values: list, dict, :ref:`IOValues<IOValues>`

        :param masks: Optional mask record for composite values. The mask definition must be included in ``ioconfig``
                It can be:

                        - A list naming the bit mask to set
                        - A dict with the mask name as a key and a boolean as the value (True to set the mask, False to clear it)
                        - An instance of :ref:`IOMask<IOMask>`
                        - A boolean value to set all masks to the same value.
        :type masks: list, dict, :ref:`IOMask<IOMask>`, bool

        :param ioconfig: Definition of DID codecs. Dictionary mapping a DID (int) to a valid :ref:`DidCodec<DidCodec>` class or pack/unpack string. 
                It is possible to use composite :ref:`DidCodec<DidCodec>` by specifying a dict with entries : codec, mask, mask_size.
        :type ioconfig: dict[int] = :ref:`DidCodec<DidCodec>`, dict

        :raises ValueError: If parameters are out of range, missing or wrong type
        :raises ConfigError: If given DID is not defined within ioconfig
        """

        tools.validate_int(did, min=0, max=0xffff, name='DID')
        if control_param is not None:
            if not isinstance(control_param, int):
                raise ValueError("control_param  must be a valid integer")

            if control_param < 0 or control_param > 3:
                raise ValueError(
                    'control_param must either be returnControlToECU(0), resetToDefault(1), freezeCurrentState(2), shortTermAdjustment(3). %d given.' % control_param)

        if values is not None:
            if isinstance(values, list):
                values = IOValues(*values)
            if isinstance(values, dict):
                values = IOValues(**values)

            if not isinstance(values, IOValues):
                raise ValueError("values must be an instance of IOValues")

        if masks is not None:
            if isinstance(masks, list):
                masks = IOMasks(*masks)
            if isinstance(masks, dict):
                masks = IOMasks(**masks)

            if not isinstance(masks, IOMasks) and not isinstance(masks, bool):
                raise ValueError("masks must be an instance of IOMask or a boolean value")

        if values is None and masks is not None:
            raise ValueError('An IOValue must be given if a IOMask is provided.')

        if ioconfig is None:
            raise ValueError("ioconfig must be provided")

        request = Request(service=cls)

        request.data = bytes()
        # IO dids are defined in client config.
        ioconfig_validated = tools.check_io_config(did, ioconfig)
        io_config_entry = tools.fetch_io_entry_from_config(did, ioconfig_validated)
        request.data += struct.pack('>H', did)

        # This parameter is optional according to standard
        if control_param is not None:
            request.data += struct.pack('B', control_param)
       # tools.
        codec = make_did_codec_from_definition(io_config_entry)  # Get IO codec from config

        if values is not None:
            request.data += codec.encode(*values.args, **values.kwargs)

        if masks is not None:  # Skip the masks byte if none is given.
            if isinstance(masks, bool):
                byte: bytes = b'\xFF' if masks == True else b'\x00'
                if 'mask_size' in io_config_entry and io_config_entry['mask_size'] is not None:
                    mask_size = io_config_entry['mask_size']
                    assert mask_size is not None    # mypy nitpick
                    request.data += (byte * mask_size)
                else:
                    raise ConfigError(
                        'mask_size', msg='Given mask is boolean value, indicating that all mask should be set to same value, but no mask_size is defined in configuration. Cannot guess how many bits to set.')

            elif isinstance(masks, IOMasks):
                if 'mask' not in io_config_entry:
                    raise ConfigError('mask', msg='Cannot apply given mask. Input/Output configuration does not define their position (and size).')
                masks_config = io_config_entry['mask']
                if masks_config is None:
                    masks_config = {}

                given_masks = masks.get_dict()

                numeric_val: int = 0
                for mask_name in given_masks:
                    if mask_name not in masks_config:
                        raise ConfigError('mask_size', msg='Cannot set mask bit for mask %s. The configuration does not define its position' % (mask_name))

                    if given_masks[mask_name] == True:
                        numeric_val |= masks_config[mask_name]

                size = math.ceil(math.log(numeric_val + 1, 2) / 8.0)
                if 'mask_size' in io_config_entry:
                    mask_size = io_config_entry['mask_size']
                    if mask_size is not None:
                        size = mask_size
                request.data += numeric_val.to_bytes(size, 'big')
        return request

    @classmethod
    def interpret_response(cls,
                           response: Response,
                           control_param: Optional[int] = None,
                           tolerate_zero_padding: bool = True,
                           ioconfig: Optional[IOConfig] = None
                           ) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`InputOutputControlByIdentifier.ResponseData<udsoncan.services.InputOutputControlByIdentifier.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :param control_param:  Same optional control parameter value given to make_request()
        :type control_param: int

        :param tolerate_zero_padding: Ignore trailing zeros in the response data avoiding raising false :class:`InvalidResponseException<udsoncan.exceptions.InvalidResponseException>`.
        :type tolerate_zero_padding: bool

        :param ioconfig: Definition of DID codecs. Dictionary mapping a DID (int) to a valid :ref:`DidCodec<DidCodec>` class or pack/unpack string. 
                It is possible to use composite DidCodec by specifying a dict with entries : codec, mask, mask_size.
        :type ioconfig: dict[int] = :ref:`DidCodec<DidCodec>`, dict

        :raises ValueError: If parameters are out of range, missing or wrong type
        :raises ConfigError: If DID echoed back by the server is not in the ``ioconfig`` definition
        :raises InvalidResponseException: If response data is incomplete or if DID data does not match codec length.
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if ioconfig is None:
            raise ValueError("IoConfig must be defined")

        min_response_size = 2 if control_param is not None else 1  # Spec specifies that if first byte is a ControlParameter, it must be echoed back by the server

        if len(response.data) < min_response_size:
            raise InvalidResponseException(response, "Response must be at least %d bytes long" % min_response_size)

        response.service_data = cls.ResponseData(
            did_echo=struct.unpack(">H", response.data[0:2])[0]
        )

        did = response.service_data.did_echo
        ioconfig_validated = tools.check_io_config(did, ioconfig)  # IO DIDs are defined in client config.
        io_config_entry = tools.fetch_io_entry_from_config(did, ioconfig_validated)  # Get requested did definition (given or default)
        codec = make_did_codec_from_definition(io_config_entry)  # Get IO codec from config

        next_byte = 2
        if control_param is not None:
            if len(response.data) < next_byte:
                raise InvalidResponseException(
                    response, 'Response should include an echo of the InputOutputControlParameter (0x%02x)' % control_param)
            response.service_data.control_param_echo = response.data[next_byte]
            next_byte += 1

        if len(response.data) >= next_byte:
            remaining_data = response.data[next_byte:]

            try:
                payload_size = len(codec)
            except DidCodec.ReadAllRemainingData:
                payload_size = len(response.data) - next_byte

            if len(remaining_data) > payload_size:
                if remaining_data[payload_size:] == b'\x00' * (len(remaining_data) - payload_size):
                    if tolerate_zero_padding:
                        remaining_data = remaining_data[0:payload_size]
            try:
                response.service_data.decoded_data = codec.decode(remaining_data)
            except Exception as e:
                raise InvalidResponseException(response, 'Response from server could not be decoded. Exception is : %s' % e)

        return cast(InputOutputControlByIdentifier.InterpretedResponse, response)



from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan.BaseService import BaseService, BaseSubfunction, BaseResponseData
import udsoncan.tools as tools
from udsoncan.ResponseCode import ResponseCode
from udsoncan.exceptions import *
import struct

from typing import Optional, cast


class RoutineControl(BaseService):
    _sid = 0x31

    class ControlType(BaseSubfunction):
        """
        RoutineControl defined subfunctions
        """
        __pretty_name__ = 'control type'

        startRoutine = 1
        stopRoutine = 2
        requestRoutineResults = 3

    supported_negative_response = [ResponseCode.SubFunctionNotSupported,
                                   ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestSequenceError,
                                   ResponseCode.RequestOutOfRange,
                                   ResponseCode.SecurityAccessDenied,
                                   ResponseCode.GeneralProgrammingFailure
                                   ]

    class ResponseData(BaseResponseData):
        """
        .. data:: control_type_echo

                Requests subfunction echoed back by the server

        .. data:: routine_id_echo

                Requests routine ID echoed back by the server.

        .. data:: routine_status_record

                Additional data associated with the response.
        """

        control_type_echo: int
        routine_id_echo: int
        routine_status_record: bytes

        def __init__(self, control_type_echo: int, routine_id_echo: int, routine_status_record: bytes):
            super().__init__(RoutineControl)

            self.control_type_echo = control_type_echo
            self.routine_id_echo = routine_id_echo
            self.routine_status_record = routine_status_record

    class InterpretedResponse(Response):
        service_data: "RoutineControl.ResponseData"

    @classmethod
    def make_request(cls, routine_id: int, control_type: int, data: Optional[bytes] = None) -> Request:
        """
        Generates a request for RoutineControl

        :param routine_id: The routine ID. Value should be between 0 and 0xFFFF
        :type routine_id: int

        :param control_type: Service subfunction. Allowed values are from 0 to 0x7F
        :type control_type: int

        :param data: Optional additional data to provide to the server
        :type data: bytes

        :raises ValueError: If parameters are out of range, missing or wrong type
        """

        tools.validate_int(routine_id, min=0, max=0xFFFF, name='Routine ID')
        tools.validate_int(control_type, min=0, max=0x7F, name='Routine control type')

        if data is not None:
            if not isinstance(data, bytes):
                raise ValueError('data must be a valid bytes object')

        request = Request(service=cls, subfunction=control_type)
        request.data = struct.pack('>H', routine_id)
        if data is not None:
            request.data += data

        return request

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`RoutineControl.ResponseData<udsoncan.services.RoutineControl.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :raises InvalidResponseException: If length of ``response.data`` is too short
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 3:
            raise InvalidResponseException(response, "Response data must be at least 3 bytes")

        response.service_data = cls.ResponseData(
            control_type_echo=response.data[0],
            routine_id_echo=struct.unpack(">H", response.data[1:3])[0],
            routine_status_record=response.data[3:] if len(response.data) > 3 else b''
        )

        return cast(RoutineControl.InterpretedResponse, response)



import struct
from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan import latest_standard
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseSubfunction, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
import udsoncan.tools as tools

from typing import Optional, cast


class DiagnosticSessionControl(BaseService):
    _sid = 0x10

    supported_negative_response = [ResponseCode.SubFunctionNotSupported,
                                   ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect
                                   ]

    class Session(BaseSubfunction):
        """
        DiagnosticSessionControl defined subfunctions
        """
        __pretty_name__ = 'session'

        defaultSession = 1
        programmingSession = 2
        extendedDiagnosticSession = 3
        safetySystemDiagnosticSession = 4

    class ResponseData(BaseResponseData):
        """
        .. data:: session_echo

                Request subfunction echoed back by the server

        .. data:: session_param_records

                Raw session parameter records. Data given by the server. For 2006 configurations, this data can is manufacturer specific. For 2013 version and above, this data correspond to P2 and P2* timing requirement.

        .. data:: p2_server_max

                Default P2 max timing supported by the server for the activated diagnostic session. Applicable for 2013 version and above. Value in seconds.

        .. data:: p2_star_server_max

                Default P2* (NRC 0x78) max timing supported by the server for the activated diagnostic session. Applicable for 2013 version and above. Value in seconds
        """
        session_echo: int
        session_param_records: bytes
        p2_server_max: Optional[float]
        p2_star_server_max: Optional[float]

        def __init__(self, session_echo: int, session_param_records: bytes, p2_server_max: Optional[float] = None, p2_star_server_max: Optional[float] = None):
            super().__init__(DiagnosticSessionControl)
            self.session_echo = session_echo
            self.session_param_records = session_param_records
            self.p2_server_max = p2_server_max
            self.p2_star_server_max = p2_star_server_max

    class InterpretedResponse(Response):
        service_data: "DiagnosticSessionControl.ResponseData"

    @classmethod
    def make_request(cls, session: int) -> Request:
        """
        Generates a request for DiagnosticSessionControl service

        :param session: Service subfunction. Allowed values are from 0 to 0x7F
        :type session: int

        :raises ValueError: If parameters are out of range, missing or wrong type
        """

        tools.validate_int(session, min=0, max=0x7F, name='Session number')
        return Request(service=cls, subfunction=session)

    @classmethod
    def interpret_response(cls, response: Response, standard_version: int = latest_standard) -> "DiagnosticSessionControl.InterpretedResponse":
        """
        Populates the response ``service_data`` property with an instance of :class:`DiagnosticSessionControl.ResponseData<udsoncan.services.DiagnosticSessionControl.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :param standard_version: The version of the ISO-14229 (the year). eg. 2006, 2013, 2020
        :type standard_version: int

        :raises InvalidResponseException: If length of ``response.data`` is too short
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 1: 	# Should not happen as response decoder will raise an exception.
            raise InvalidResponseException(response, "Response data must be at least 1 bytes")

        response.service_data = cls.ResponseData(
            session_echo=response.data[0],
            session_param_records=response.data[1:] if len(response.data) > 1 else b''
        )

        if (standard_version >= 2013):
            if len(response.data) != 5:
                raise InvalidResponseException(
                    response, 'Response must contain 4 bytes of data representing the server timing requirements (P2 and P2* timeouts). Got %d bytes' % len(response.data))

            (a, b) = struct.unpack('>HH', response.data[1:])
            response.service_data.p2_server_max = (a) / 1000
            response.service_data.p2_star_server_max = (b * 10) / 1000

        return cast(DiagnosticSessionControl.InterpretedResponse, response)



import struct

__all__ = ['AddressAndLengthFormatIdentifier']


class AddressAndLengthFormatIdentifier:
    """
    This class defines how many bytes of a memorylocation, composed of an address and a memorysize, should be encoded when sent over the underlying protocol.
    Mainly used by :ref:`ReadMemoryByAddress<ReadMemoryByAddress>`, :ref:`WriteMemoryByAddress<WriteMemoryByAddress>`, :ref:`RequestDownload<RequestDownload>` and :ref:`RequestUpload<RequestUpload>` services

    Defined by ISO-14229:2020, Annex H

    :param address_format: The number of bits on which an address should be encoded. Possible values are 8, 16, 24, 32, 40, 48, 56, 64
    :type address_format: int

    :param memorysize_format: The number of bits on which a memory size should be encoded. Possible values are 8, 16, 24, 32, 40, 48, 56, 64
    :type memorysize_format: int

    """
    address_map = {
        8: 1,
        16: 2,
        24: 3,
        32: 4,
        40: 5,
        48: 6,
        56: 7,
        64: 8
    }

    memsize_map = {
        8: 1,
        16: 2,
        24: 3,
        32: 4,
        40: 5,
        48: 6,
        56: 7,
        64: 8
    }

    memorysize_format: int
    address_format: int

    def __init__(self, address_format: int, memorysize_format: int):
        if address_format not in self.address_map:
            raise ValueError('address_format must ba an integer selected from : %s ' % (self.address_map.keys()))

        if not isinstance(memorysize_format, int) or not isinstance(address_format, int):
            raise ValueError('memorysize_format and address_format must be integers')

        if memorysize_format not in self.memsize_map:
            raise ValueError('memorysize_format must be an integer selected from : %s' % (self.memsize_map.keys()))

        self.memorysize_format = memorysize_format
        self.address_format = address_format

    def get_byte_as_int(self) -> int:
        return ((self.memsize_map[self.memorysize_format] << 4) | (self.address_map[self.address_format])) & 0xFF

    # Byte given alongside a memory address and a length so that they are decoded properly.
    def get_byte(self) -> bytes:
        return struct.pack('B', self.get_byte_as_int())



from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseResponseData
from udsoncan.ResponseCode import ResponseCode


class ReadDataByPeriodicIdentifier(BaseService):
    _sid = 0x2A

    supported_negative_response = [ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestOutOfRange,
                                   ResponseCode.SecurityAccessDenied
                                   ]

    class ResponseData(BaseResponseData):
        def __init__(self):
            super().__init__(ReadDataByPeriodicIdentifier)

    class InterpretedResponse(Response):
        service_data: "ReadDataByPeriodicIdentifier.ResponseData"

    @classmethod
    def make_request(cls) -> Request:
        raise NotImplementedError('Service is not implemented')

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        raise NotImplementedError('Service is not implemented')



from udsoncan import DataFormatIdentifier, AddressAndLengthFormatIdentifier, MemoryLocation, CommunicationType, Baudrate, IOMasks, Dtc, DidCodec, AsciiCodec, Filesize, DynamicDidDefinition, make_did_codec_from_definition
from test.UdsTest import UdsTest
import struct


class TestAddressAndLengthFormatIdentifier(UdsTest):
    def test_ali_1(self):
        alfid = AddressAndLengthFormatIdentifier(memorysize_format=8, address_format=8)
        self.assertEqual(alfid.get_byte(), b'\x11')

    def test_ali_2(self):
        alfid = AddressAndLengthFormatIdentifier(memorysize_format=16, address_format=8)
        self.assertEqual(alfid.get_byte(), b'\x21')

    def test_ali_oob_values(self):  # Out Of Bounds Value
        with self.assertRaises(ValueError):
            AddressAndLengthFormatIdentifier(memorysize_format=1, address_format=1)

        with self.assertRaises(ValueError):
            AddressAndLengthFormatIdentifier(memorysize_format=0, address_format=8)

        with self.assertRaises(ValueError):
            AddressAndLengthFormatIdentifier(memorysize_format=8, address_format=0)

        with self.assertRaises(ValueError):
            AddressAndLengthFormatIdentifier(memorysize_format=40, address_format=0)

        with self.assertRaises(ValueError):
            AddressAndLengthFormatIdentifier(memorysize_format=8, address_format=65)

        with self.assertRaises(ValueError):
            AddressAndLengthFormatIdentifier(memorysize_format='8', address_format=8)

        with self.assertRaises(ValueError):
            AddressAndLengthFormatIdentifier(memorysize_format=8, address_format='8')

    def test_str_repr(self):
        alfid = AddressAndLengthFormatIdentifier(memorysize_format=8, address_format=8)
        str(alfid)
        alfid.__repr__()


class TestDataFormatIdentifier(UdsTest):
    def test_dfi(self):
        dfi = DataFormatIdentifier(compression=1, encryption=2)
        self.assertEqual(dfi.get_byte(), b'\x12')

    def test_dfi2(self):
        dfi = DataFormatIdentifier(compression=15, encryption=15)
        self.assertEqual(dfi.get_byte(), b'\xFF')

    def test_str_repr(self):
        dfi = DataFormatIdentifier(compression=1, encryption=2)
        str(dfi)
        dfi.__repr__()

    def test_from_byte(self):
        dfi = DataFormatIdentifier.from_byte(0xAB)
        self.assertEqual(dfi.compression, 0xA)
        self.assertEqual(dfi.encryption, 0xB)

    def test_ali_oob_values(self):
        with self.assertRaises(ValueError):
            DataFormatIdentifier(compression=-1, encryption=1)

        with self.assertRaises(ValueError):
            DataFormatIdentifier(compression=1, encryption=-1)

        with self.assertRaises(ValueError):
            DataFormatIdentifier(compression=16, encryption=1)

        with self.assertRaises(ValueError):
            DataFormatIdentifier(compression=1, encryption=16)


class TestMemoryLocation(UdsTest):
    def test_memloc1(self):
        memloc = MemoryLocation(address=0x1234, memorysize=0x78, address_format=16, memorysize_format=8)
        self.assertEqual(memloc.get_address_bytes(), b'\x12\x34')
        self.assertEqual(memloc.get_memorysize_bytes(), b'\x78')

    def test_memloc_autosize1(self):
        memloc = MemoryLocation(address=0x1234, memorysize=0x78)
        self.assertEqual(memloc.get_address_bytes(), b'\x12\x34')
        self.assertEqual(memloc.get_memorysize_bytes(), b'\x78')

    def test_memloc_autosize2(self):
        memloc = MemoryLocation(address=0x1234567, memorysize=0x789abb)
        self.assertEqual(memloc.get_address_bytes(), b'\x01\x23\x45\x67')
        self.assertEqual(memloc.get_memorysize_bytes(), b'\x78\x9a\xbb')

    def test_memloc_str_repr(self):
        memloc = MemoryLocation(address=0x1234, memorysize=0x78, address_format=16, memorysize_format=8)
        str(memloc)
        memloc.__repr__()

    def test_memloc_override(self):
        memloc = MemoryLocation(address=0x1234, memorysize=0x78)
        self.assertEqual(memloc.get_address_bytes(), b'\x12\x34')
        self.assertEqual(memloc.get_memorysize_bytes(), b'\x78')
        memloc.set_format_if_none(address_format=32)
        self.assertEqual(memloc.get_address_bytes(), b'\x00\x00\x12\x34')
        self.assertEqual(memloc.get_memorysize_bytes(), b'\x78')
        memloc.set_format_if_none(memorysize_format=24)
        self.assertEqual(memloc.get_address_bytes(), b'\x00\x00\x12\x34')
        self.assertEqual(memloc.get_memorysize_bytes(), b'\x00\x00\x78')

        memloc = MemoryLocation(address=0x1234, memorysize=0x78)
        memloc.set_format_if_none(address_format=32, memorysize_format=24)  # Both at same time.
        self.assertEqual(memloc.get_address_bytes(), b'\x00\x00\x12\x34')
        self.assertEqual(memloc.get_memorysize_bytes(), b'\x00\x00\x78')

    def test_memloc_from_bytes(self):
        memloc = MemoryLocation.from_bytes(address_bytes=b'\x12\x34', memorysize_bytes=b'\xFF')
        self.assertEqual(memloc.address, 0x1234)
        self.assertEqual(memloc.memorysize, 0xFF)
        self.assertEqual(memloc.address_format, 16)
        self.assertEqual(memloc.memorysize_format, 8)

        memloc = MemoryLocation.from_bytes(address_bytes=b'\x12\x34\x56', memorysize_bytes=b'\x66\x77\x88')
        self.assertEqual(memloc.address, 0x123456)
        self.assertEqual(memloc.memorysize, 0x667788)
        self.assertEqual(memloc.address_format, 24)
        self.assertEqual(memloc.memorysize_format, 24)

    def test_memloc_max_size(self):
        MemoryLocation.from_bytes(address_bytes=b'\x12\x34\x56\x78\x9a', memorysize_bytes=b'\xFF')
        with self.assertRaises(ValueError):
            MemoryLocation.from_bytes(address_bytes=b'\x12\x34\x56\x78\x9a\xbc', memorysize_bytes=b'\xFF')

        MemoryLocation.from_bytes(address_bytes=b'\x12\x34', memorysize_bytes=b'\x12\x34\x56\x78')
        with self.assertRaises(ValueError):
            MemoryLocation.from_bytes(address_bytes=b'\x12\x34', memorysize_bytes=b'\x12\x34\x56\x78\x9a')


class TestCommunicationType(UdsTest):
    def test_make(self):
        comtype = CommunicationType(subnet=CommunicationType.Subnet.node, normal_msg=True, network_management_msg=False)
        self.assertEqual(comtype.get_byte(), b'\x01')

        comtype = CommunicationType(subnet=CommunicationType.Subnet.network, normal_msg=True, network_management_msg=False)
        self.assertEqual(comtype.get_byte(), b'\xF1')

        comtype = CommunicationType(subnet=3, normal_msg=True, network_management_msg=True)
        self.assertEqual(comtype.get_byte(), b'\x33')

    def test_str_repr(self):
        comtype = CommunicationType(subnet=CommunicationType.Subnet.node, normal_msg=True, network_management_msg=False)
        str(comtype)
        comtype.__repr__()

    def test_from_byte(self):
        comtype = CommunicationType.from_byte(b'\x01')
        self.assertEqual(comtype.get_byte(), b'\x01')

        comtype = CommunicationType.from_byte(b'\xF1')
        self.assertEqual(comtype.get_byte(), b'\xF1')

        comtype = CommunicationType.from_byte(b'\x33')
        self.assertEqual(comtype.get_byte(), b'\x33')

        comtype = CommunicationType.from_byte(0x01)
        self.assertEqual(comtype.get_byte(), b'\x01')

    def test_oob_values(self):
        with self.assertRaises(ValueError):
            CommunicationType(subnet=0, normal_msg=False, network_management_msg=False)

        with self.assertRaises(ValueError):
            CommunicationType(subnet='x', normal_msg=True, network_management_msg=False)

        with self.assertRaises(ValueError):
            CommunicationType(subnet=0, normal_msg=1, network_management_msg=True)

        with self.assertRaises(ValueError):
            CommunicationType(subnet=0, normal_msg=True, network_management_msg=1)


class TestBaudrate(UdsTest):
    def test_create_fixed(self):
        br = Baudrate(115200, baudtype=Baudrate.Type.Fixed)
        self.assertEqual(br.get_bytes(), b'\x05')

        with self.assertRaises(ValueError):
            br = Baudrate(123456, baudtype=Baudrate.Type.Fixed)

    def test_create_specific(self):
        br = Baudrate(115200, baudtype=Baudrate.Type.Specific)
        self.assertEqual(br.get_bytes(), b'\x01\xC2\x00')

        with self.assertRaises(ValueError):
            br = Baudrate(0x1000000, baudtype=Baudrate.Type.Specific)

    def test_create_id(self):
        for i in range(0xFF):
            br = Baudrate(i, baudtype=Baudrate.Type.Identifier)
            self.assertEqual(br.get_bytes(), struct.pack('B', i))

        with self.assertRaises(ValueError):
            br = Baudrate(0x100, baudtype=Baudrate.Type.Identifier)

    def test_effective_baudrate(self):
        br = Baudrate(0x12, Baudrate.Type.Identifier)  # 500kbits
        self.assertEqual(br.effective_baudrate(), 500000)

    def test_change_type(self):
        br = Baudrate(115200, baudtype=Baudrate.Type.Fixed)
        br2 = br.make_new_type(Baudrate.Type.Specific)
        self.assertEqual(br2.get_bytes(), b'\x01\xC2\x00')

        br = Baudrate(115200, baudtype=Baudrate.Type.Specific)
        br2 = br.make_new_type(Baudrate.Type.Fixed)
        self.assertEqual(br2.get_bytes(), b'\x05')

    def test_str_repr(self):
        br = Baudrate(115200, baudtype=Baudrate.Type.Fixed)
        str(br)
        br.__repr__()

    def test_create_auto(self):
        # Direct ID
        br = Baudrate(1)
        self.assertEqual(br.get_bytes(), b'\x01')

        br = Baudrate(0xFF)
        self.assertEqual(br.get_bytes(), b'\xFF')

        # Fixed baudrate
        br = Baudrate(115200)
        self.assertEqual(br.get_bytes(), b'\x05')

        br = Baudrate(500000)
        self.assertEqual(br.get_bytes(), b'\x12')

        # Specific Baudrate:
        br = Baudrate(0x123456)
        self.assertEqual(br.get_bytes(), b'\x12\x34\x56')

    def test_oob_values(self):
        with self.assertRaises(ValueError):
            br = Baudrate(-1)

        with self.assertRaises(ValueError):
            br = Baudrate(1, baudtype=-1)

        with self.assertRaises(ValueError):
            br = Baudrate(1, baudtype=0xFF)


class TestIOMasks(UdsTest):
    def test_oob_values(self):
        with self.assertRaises(ValueError):
            IOMasks(aaa='asd')

        with self.assertRaises(ValueError):
            IOMasks(1, 2, 3)

    def test_make_dict(self):
        m = IOMasks('aaa', 'bbb')  # Correct syntax
        self.assertEqual(m.get_dict(), {'aaa': True, 'bbb': True})

        m = IOMasks('aaa', 'bbb', ccc=True, ddd=False)  # Correct syntax
        self.assertEqual(m.get_dict(), {'aaa': True, 'bbb': True, 'ccc': True, 'ddd': False})


class TestDtc(UdsTest):
    def test_init(self):
        dtc = Dtc(0x1234)
        self.assertEqual(dtc.id, 0x1234)
        self.assertEqual(dtc.status.get_byte(), b'\x00')
        self.assertEqual(dtc.status.get_byte_as_int(), 0x00)
        self.assertEqual(dtc.severity.get_byte(), b'\x00')
        self.assertEqual(dtc.severity.get_byte_as_int(), 0x00)

        self.assertEqual(dtc.status.test_failed, False)
        self.assertEqual(dtc.status.test_failed_this_operation_cycle, False)
        self.assertEqual(dtc.status.pending, False)
        self.assertEqual(dtc.status.confirmed, False)
        self.assertEqual(dtc.status.test_not_completed_since_last_clear, False)
        self.assertEqual(dtc.status.test_failed_since_last_clear, False)
        self.assertEqual(dtc.status.test_not_completed_this_operation_cycle, False)
        self.assertEqual(dtc.status.warning_indicator_requested, False)

        with self.assertRaises(TypeError):
            Dtc()

    def test_set_status_with_byte_no_error(self):
        dtc = Dtc(1)
        for i in range(255):
            dtc.status.set_byte(i)

    def test_status_behaviour(self):
        dtc = Dtc(1)

        self.assertEqual(dtc.status.get_byte(), b'\x00')
        dtc.status.test_failed = True
        self.assertEqual(dtc.status.get_byte(), b'\x01')
        dtc.status.test_failed_this_operation_cycle = True
        self.assertEqual(dtc.status.get_byte(), b'\x03')
        dtc.status.pending = True
        self.assertEqual(dtc.status.get_byte(), b'\x07')
        dtc.status.confirmed = True
        self.assertEqual(dtc.status.get_byte(), b'\x0F')
        dtc.status.test_not_completed_since_last_clear = True
        self.assertEqual(dtc.status.get_byte(), b'\x1F')
        dtc.status.test_failed_since_last_clear = True
        self.assertEqual(dtc.status.get_byte(), b'\x3F')
        dtc.status.test_not_completed_this_operation_cycle = True
        self.assertEqual(dtc.status.get_byte(), b'\x7F')
        dtc.status.warning_indicator_requested = True
        self.assertEqual(dtc.status.get_byte(), b'\xFF')

        dtc.status.set_byte(0x01)
        self.assertEqual(dtc.status.test_failed, True)
        self.assertEqual(dtc.status.test_failed_this_operation_cycle, False)
        self.assertEqual(dtc.status.pending, False)
        self.assertEqual(dtc.status.confirmed, False)
        self.assertEqual(dtc.status.test_not_completed_since_last_clear, False)
        self.assertEqual(dtc.status.test_failed_since_last_clear, False)
        self.assertEqual(dtc.status.test_not_completed_this_operation_cycle, False)
        self.assertEqual(dtc.status.warning_indicator_requested, False)

        dtc.status.set_byte(0x03)
        self.assertEqual(dtc.status.test_failed, True)
        self.assertEqual(dtc.status.test_failed_this_operation_cycle, True)
        self.assertEqual(dtc.status.pending, False)
        self.assertEqual(dtc.status.confirmed, False)
        self.assertEqual(dtc.status.test_not_completed_since_last_clear, False)
        self.assertEqual(dtc.status.test_failed_since_last_clear, False)
        self.assertEqual(dtc.status.test_not_completed_this_operation_cycle, False)
        self.assertEqual(dtc.status.warning_indicator_requested, False)

        dtc.status.set_byte(0x07)
        self.assertEqual(dtc.status.test_failed, True)
        self.assertEqual(dtc.status.test_failed_this_operation_cycle, True)
        self.assertEqual(dtc.status.pending, True)
        self.assertEqual(dtc.status.confirmed, False)
        self.assertEqual(dtc.status.test_not_completed_since_last_clear, False)
        self.assertEqual(dtc.status.test_failed_since_last_clear, False)
        self.assertEqual(dtc.status.test_not_completed_this_operation_cycle, False)
        self.assertEqual(dtc.status.warning_indicator_requested, False)

        dtc.status.set_byte(0x0F)
        self.assertEqual(dtc.status.test_failed, True)
        self.assertEqual(dtc.status.test_failed_this_operation_cycle, True)
        self.assertEqual(dtc.status.pending, True)
        self.assertEqual(dtc.status.confirmed, True)
        self.assertEqual(dtc.status.test_not_completed_since_last_clear, False)
        self.assertEqual(dtc.status.test_failed_since_last_clear, False)
        self.assertEqual(dtc.status.test_not_completed_this_operation_cycle, False)
        self.assertEqual(dtc.status.warning_indicator_requested, False)

        dtc.status.set_byte(0x1F)
        self.assertEqual(dtc.status.test_failed, True)
        self.assertEqual(dtc.status.test_failed_this_operation_cycle, True)
        self.assertEqual(dtc.status.pending, True)
        self.assertEqual(dtc.status.confirmed, True)
        self.assertEqual(dtc.status.test_not_completed_since_last_clear, True)
        self.assertEqual(dtc.status.test_failed_since_last_clear, False)
        self.assertEqual(dtc.status.test_not_completed_this_operation_cycle, False)
        self.assertEqual(dtc.status.warning_indicator_requested, False)

        dtc.status.set_byte(0x3F)
        self.assertEqual(dtc.status.test_failed, True)
        self.assertEqual(dtc.status.test_failed_this_operation_cycle, True)
        self.assertEqual(dtc.status.pending, True)
        self.assertEqual(dtc.status.confirmed, True)
        self.assertEqual(dtc.status.test_not_completed_since_last_clear, True)
        self.assertEqual(dtc.status.test_failed_since_last_clear, True)
        self.assertEqual(dtc.status.test_not_completed_this_operation_cycle, False)
        self.assertEqual(dtc.status.warning_indicator_requested, False)

        dtc.status.set_byte(0x7F)
        self.assertEqual(dtc.status.test_failed, True)
        self.assertEqual(dtc.status.test_failed_this_operation_cycle, True)
        self.assertEqual(dtc.status.pending, True)
        self.assertEqual(dtc.status.confirmed, True)
        self.assertEqual(dtc.status.test_not_completed_since_last_clear, True)
        self.assertEqual(dtc.status.test_failed_since_last_clear, True)
        self.assertEqual(dtc.status.test_not_completed_this_operation_cycle, True)
        self.assertEqual(dtc.status.warning_indicator_requested, False)

        dtc.status.set_byte(0xFF)
        self.assertEqual(dtc.status.test_failed, True)
        self.assertEqual(dtc.status.test_failed_this_operation_cycle, True)
        self.assertEqual(dtc.status.pending, True)
        self.assertEqual(dtc.status.confirmed, True)
        self.assertEqual(dtc.status.test_not_completed_since_last_clear, True)
        self.assertEqual(dtc.status.test_failed_since_last_clear, True)
        self.assertEqual(dtc.status.test_not_completed_this_operation_cycle, True)
        self.assertEqual(dtc.status.warning_indicator_requested, True)

    def test_set_severity_with_byte_no_error(self):
        dtc = Dtc(1)
        for i in range(255):
            dtc.severity.set_byte(i)

    def test_str_repr(self):
        dtc = Dtc(0x123456)
        dtc.status.pending = True
        str(dtc)
        dtc.__repr__()

    def test_severity_behaviour(self):
        dtc = Dtc(1)

        self.assertEqual(dtc.severity.get_byte_as_int(), 0x00)
        dtc.severity.maintenance_only = True
        self.assertEqual(dtc.severity.get_byte_as_int(), 0x20)
        dtc.severity.check_at_next_exit = True
        self.assertEqual(dtc.severity.get_byte_as_int(), 0x60)
        dtc.severity.check_immediately = True
        self.assertEqual(dtc.severity.get_byte_as_int(), 0xE0)

        dtc.severity.set_byte(0x20)
        self.assertEqual(dtc.severity.maintenance_only, True)
        self.assertEqual(dtc.severity.check_at_next_exit, False)
        self.assertEqual(dtc.severity.check_immediately, False)

        dtc.severity.set_byte(0x60)
        self.assertEqual(dtc.severity.maintenance_only, True)
        self.assertEqual(dtc.severity.check_at_next_exit, True)
        self.assertEqual(dtc.severity.check_immediately, False)

        dtc.severity.set_byte(0xE0)
        self.assertEqual(dtc.severity.maintenance_only, True)
        self.assertEqual(dtc.severity.check_at_next_exit, True)
        self.assertEqual(dtc.severity.check_immediately, True)


class TestCodec(UdsTest):
    def test_DIDCodec_bad_values(self):
        with self.assertRaises(NotImplementedError):
            codec = DidCodec()
            codec.encode("asd")

        with self.assertRaises(NotImplementedError):
            codec = DidCodec()
            codec.decode(b"asd")

        with self.assertRaises(ValueError):
            make_did_codec_from_definition("")

    def test_ascii_codec(self):
        codec = AsciiCodec(10)
        self.assertEqual(codec.encode("abcdefghij"), b'abcdefghij');
        self.assertEqual(codec.decode(b"abcdefghij"), 'abcdefghij');

        with self.assertRaises(ValueError):
            codec.encode("abc")

        with self.assertRaises(ValueError):
            codec.encode("abcdefghijklmnop")

        with self.assertRaises(Exception):
            AsciiCodec()


class TestFilesize(UdsTest):

    def test_create(self):
        # Normal use case
        Filesize(uncompressed=123)
        Filesize(compressed=123)
        Filesize(uncompressed=123, compressed=100)
        Filesize(uncompressed=123, compressed=100, width=2)
        Filesize(uncompressed=123, width=2)
        Filesize(compressed=100, width=2)

    def test_width(self):

        self.assertEqual(Filesize(uncompressed=0xFF).get_width(), 1)
        self.assertEqual(Filesize(uncompressed=0x100).get_width(), 2)
        self.assertEqual(Filesize(uncompressed=0xFFFF).get_width(), 2)
        self.assertEqual(Filesize(uncompressed=0x10000).get_width(), 3)
        self.assertEqual(Filesize(uncompressed=0xFFFFFF).get_width(), 3)
        self.assertEqual(Filesize(uncompressed=0x1000000).get_width(), 4)
        self.assertEqual(Filesize(uncompressed=0xFFFFFFFF).get_width(), 4)

        self.assertEqual(Filesize(compressed=0xFF).get_width(), 1)
        self.assertEqual(Filesize(compressed=0x100).get_width(), 2)
        self.assertEqual(Filesize(compressed=0xFFFF).get_width(), 2)
        self.assertEqual(Filesize(compressed=0x10000).get_width(), 3)
        self.assertEqual(Filesize(compressed=0xFFFFFF).get_width(), 3)
        self.assertEqual(Filesize(compressed=0x1000000).get_width(), 4)
        self.assertEqual(Filesize(compressed=0xFFFFFFFF).get_width(), 4)

        self.assertEqual(Filesize(uncompressed=0xFF, compressed=0xFF).get_width(), 1)
        self.assertEqual(Filesize(uncompressed=0x100, compressed=0xFF).get_width(), 2)
        self.assertEqual(Filesize(uncompressed=0xFF, compressed=0x100).get_width(), 2)
        self.assertEqual(Filesize(uncompressed=0xFFFFFF, compressed=0x100).get_width(), 3)
        self.assertEqual(Filesize(uncompressed=0xFFFFFF, compressed=0xFFFFFFFF).get_width(), 4)

        self.assertEqual(Filesize(uncompressed=0xFF, compressed=0xFF, width=4).get_width(), 4)
        self.assertEqual(Filesize(uncompressed=0xFF, compressed=0xFF, width=8).get_width(), 8)

        with self.assertRaises(ValueError):
            Filesize(uncompressed=0x100, compressed=0x100, width=1)

        with self.assertRaises(ValueError):
            Filesize(uncompressed=0xFF, compressed=0x100, width=1)

        with self.assertRaises(ValueError):
            Filesize(uncompressed=0x100, compressed=0xFF, width=1)

    def test_bytes(self):

        self.assertEqual(Filesize(uncompressed=0xFF).get_uncompressed_bytes(), b'\xFF')
        self.assertEqual(Filesize(uncompressed=0x12345678).get_uncompressed_bytes(), b'\x12\x34\x56\x78')
        self.assertEqual(Filesize(uncompressed=0x1234, width=4).get_uncompressed_bytes(), b'\x00\x00\x12\x34')
        self.assertEqual(Filesize(uncompressed=0xFF).get_compressed_bytes(), b'')

        self.assertEqual(Filesize(compressed=0xFF).get_compressed_bytes(), b'\xFF')
        self.assertEqual(Filesize(compressed=0x12345678).get_compressed_bytes(), b'\x12\x34\x56\x78')
        self.assertEqual(Filesize(compressed=0x1234, width=4).get_compressed_bytes(), b'\x00\x00\x12\x34')
        self.assertEqual(Filesize(compressed=0x12345678).get_uncompressed_bytes(), b'')

    def test_bad_values(self):
        with self.assertRaises(ValueError):
            Filesize();

        with self.assertRaises(ValueError):
            Filesize(uncompressed='asd')

        with self.assertRaises(ValueError):
            Filesize(compressed='asd')

        with self.assertRaises(ValueError):
            Filesize(compressed=123, width='asd')

        with self.assertRaises(ValueError):
            Filesize(compressed=123, width=0)

    def test_str_repr(self):
        fs = Filesize(uncompressed=123)
        str(fs)
        fs.__repr__()


class TestDynamicDidDefinition(UdsTest):
    def test_def_mismatch(self):
        diddef = DynamicDidDefinition()
        diddef.add(source_did=0x1234, position=1, memorysize=1)
        with self.assertRaises(ValueError):
            diddef.add(MemoryLocation(address=0x1234, memorysize=1))

    def test_type(self):
        diddef = DynamicDidDefinition()

        self.assertFalse(diddef.is_by_source_did())
        self.assertFalse(diddef.is_by_memory_address())
        diddef.add(source_did=0x1234, position=1, memorysize=1)
        diddef.add(source_did=0x1234, position=2, memorysize=1)

        self.assertTrue(diddef.is_by_source_did())
        self.assertFalse(diddef.is_by_memory_address())

        diddef = DynamicDidDefinition()
        diddef.add(MemoryLocation(address=0x1234, memorysize=1))
        diddef.add(MemoryLocation(address=0x1235, memorysize=1))

        self.assertFalse(diddef.is_by_source_did())
        self.assertTrue(diddef.is_by_memory_address())

    def test_get_alfid(self):
        diddef = DynamicDidDefinition()
        diddef.add(source_did=0x1234, position=1, memorysize=1)
        diddef.add(source_did=0x1234, position=2, memorysize=1)
        with self.assertRaises(ValueError):
            diddef.get_alfid()

        diddef = DynamicDidDefinition()
        diddef.add(MemoryLocation(address=0x1234, memorysize=1, address_format=16, memorysize_format=8))
        diddef.add(MemoryLocation(address=0x1235, memorysize=1, address_format=16, memorysize_format=8))
        alfid = diddef.get_alfid()
        self.assertEqual(alfid.get_byte_as_int(), 0x12)

        diddef = DynamicDidDefinition()
        diddef.add(MemoryLocation(address=0x1234, memorysize=1, address_format=16, memorysize_format=8))
        diddef.add(MemoryLocation(address=0x1235, memorysize=1, address_format=16, memorysize_format=16))
        with self.assertRaises(ValueError):
            alfid = diddef.get_alfid()



import struct
from udsoncan.Request import Request
from udsoncan.Response import Response
from udsoncan import DidCodec, check_did_config, make_did_codec_from_definition, fetch_codec_definition_from_config, DIDConfig
from udsoncan.exceptions import *
from udsoncan.BaseService import BaseService, BaseResponseData
from udsoncan.ResponseCode import ResponseCode
import udsoncan.tools as tools

from typing import Any, cast


class WriteDataByIdentifier(BaseService):
    _sid = 0x2E
    _use_subfunction = False

    supported_negative_response = [ResponseCode.IncorrectMessageLengthOrInvalidFormat,
                                   ResponseCode.ConditionsNotCorrect,
                                   ResponseCode.RequestOutOfRange,
                                   ResponseCode.SecurityAccessDenied,
                                   ResponseCode.GeneralProgrammingFailure
                                   ]

    class ResponseData(BaseResponseData):
        """
        .. data:: did_echo

                The DID echoed back by the server
        """

        did_echo: int

        def __init__(self, did_echo: int):
            super().__init__(WriteDataByIdentifier)

            self.did_echo = did_echo

    class InterpretedResponse(Response):
        service_data: "WriteDataByIdentifier.ResponseData"

    @classmethod
    def make_request(cls, did: int, value: Any, didconfig: DIDConfig) -> Request:
        """
        Generates a request for WriteDataByIdentifier

        :param did: The data identifier to write
        :type did: int

        :param value: Value given to the :ref:`DidCodec <DidCodec>`.encode method. If involved codec is defined with a pack string (default codec), multiple values may be passed with a tuple.
        :type value: object

        :param didconfig: Definition of DID codecs. Dictionary mapping a DID (int) to a valid :ref:`DidCodec <DidCodec>` class or pack/unpack string 
        :type didconfig: dict[int] = :ref:`DidCodec <DidCodec>`

        :raises ValueError: If parameters are out of range, missing or wrong type
        :raises ConfigError: If ``didlist`` contains a DID not defined in ``didconfig``
        """

        tools.validate_int(did, min=0, max=0xFFFF, name='Data Identifier')
        req = Request(cls)
        didconfig = check_did_config(did, didconfig=didconfig)  # Make sure all DIDs are correctly defined in client config
        codec_definition = fetch_codec_definition_from_config(did, didconfig)
        req.data = struct.pack('>H', did)  # encode DID number
        codec = make_did_codec_from_definition(codec_definition)
        if codec.__class__ == DidCodec and isinstance(value, tuple):
            req.data += codec.encode(*value)    # Fixes issue #29
        else:
            req.data += codec.encode(value)

        return req

    @classmethod
    def interpret_response(cls, response: Response) -> InterpretedResponse:
        """
        Populates the response ``service_data`` property with an instance of :class:`WriteDataByIdentifier.ResponseData<udsoncan.services.WriteDataByIdentifier.ResponseData>`

        :param response: The received response to interpret
        :type response: :ref:`Response<Response>`

        :raises InvalidResponseException: If length of ``response.data`` is too short
        """
        if response.data is None:
            raise InvalidResponseException(response, "No data in response")

        if len(response.data) < 2:
            raise InvalidResponseException(response, "Response must be at least 2 bytes long")

        response.service_data = cls.ResponseData(
            did_echo=struct.unpack(">H", response.data[0:2])[0]
        )

        return cast(WriteDataByIdentifier.InterpretedResponse, response)



import inspect
from typing import Dict


class ResponseCode:
    PositiveResponse = 0
    GeneralReject = 0x10
    ServiceNotSupported = 0x11
    SubFunctionNotSupported = 0x12
    IncorrectMessageLengthOrInvalidFormat = 0x13
    ResponseTooLong = 0x14
    BusyRepeatRequest = 0x21
    ConditionsNotCorrect = 0x22
    RequestSequenceError = 0x24
    NoResponseFromSubnetComponent = 0x25
    FailurePreventsExecutionOfRequestedAction = 0x26
    RequestOutOfRange = 0x31
    SecurityAccessDenied = 0x33
    AuthenticationRequired = 0x34
    InvalidKey = 0x35
    ExceedNumberOfAttempts = 0x36
    RequiredTimeDelayNotExpired = 0x37
    SecureDataTransmissionRequired = 0x38
    SecureDataTransmissionNotAllowed = 0x39
    SecureDataVerificationFailed = 0x3A
    CertificateVerificationFailed_InvalidTimePeriod = 0x50
    CertificateVerificationFailed_InvalidSignature = 0x51
    CertificateVerificationFailed_InvalidChainOfTrust = 0x52
    CertificateVerificationFailed_InvalidType = 0x53
    CertificateVerificationFailed_InvalidFormat = 0x54
    CertificateVerificationFailed_InvalidContent = 0x55
    CertificateVerificationFailed_InvalidScope = 0x56
    CertificateVerificationFailed_InvalidCertificate = 0x57
    OwnershipVerificationFailed = 0x58
    ChallengeCalculationFailed = 0x59
    SettingAccessRightsFailed = 0x5A
    SessionKeyCreationDerivationFailed = 0x5B
    ConfigurationDataUsageFailed = 0x5C
    DeAuthenticationFailed = 0x5D
    UploadDownloadNotAccepted = 0x70
    TransferDataSuspended = 0x71
    GeneralProgrammingFailure = 0x72
    WrongBlockSequenceCounter = 0x73
    RequestCorrectlyReceived_ResponsePending = 0x78
    SubFunctionNotSupportedInActiveSession = 0x7E
    ServiceNotSupportedInActiveSession = 0x7F
    RpmTooHigh = 0x81
    RpmTooLow = 0x82
    EngineIsRunning = 0x83
    EngineIsNotRunning = 0x84
    EngineRunTimeTooLow = 0x85
    TemperatureTooHigh = 0x86
    TemperatureTooLow = 0x87
    VehicleSpeedTooHigh = 0x88
    VehicleSpeedTooLow = 0x89
    ThrottlePedalTooHigh = 0x8A
    ThrottlePedalTooLow = 0x8B
    TransmissionRangeNotInNeutral = 0x8C
    TransmissionRangeNotInGear = 0x8D
    BrakeSwitchNotClosed = 0x8F
    ShifterLeverNotInPark = 0x90
    TorqueConverterClutchLocked = 0x91
    VoltageTooHigh = 0x92
    VoltageTooLow = 0x93
    ResourceTemporarilyNotAvailable = 0x94

    # Defined by ISO-15764. Offset of 0x38 is defined within UDS standard (ISO-14229)
    GeneralSecurityViolation = 0x38 + 0
    SecuredModeRequested = 0x38 + 1
    InsufficientProtection = 0x38 + 2
    TerminationWithSignatureRequested = 0x38 + 3
    AccessDenied = 0x38 + 4
    VersionNotSupported = 0x38 + 5
    SecuredLinkNotSupported = 0x38 + 6
    CertificateNotAvailable = 0x38 + 7
    AuditTrailInformationNotAvailable = 0x38 + 8

    @classmethod
    def is_supported_by_standard(cls, code: int, standard_version: int) -> bool:
        if not isinstance(code, int):
            raise ValueError("given code must be an integer value, not %s" % (code.__class__.__name__))

        if not isinstance(standard_version, int):
            raise ValueError("given standard_version must be an integer value, not %s" % (standard_version.__class__.__name__))

        codes_version: Dict[int, int] = {
            cls.PositiveResponse: 2006,
            cls.GeneralReject: 2006,
            cls.ServiceNotSupported: 2006,
            cls.SubFunctionNotSupported: 2006,
            cls.IncorrectMessageLengthOrInvalidFormat: 2006,
            cls.ResponseTooLong: 2006,
            cls.BusyRepeatRequest: 2006,
            cls.ConditionsNotCorrect: 2006,
            cls.RequestSequenceError: 2006,
            cls.NoResponseFromSubnetComponent: 2006,
            cls.FailurePreventsExecutionOfRequestedAction: 2006,
            cls.RequestOutOfRange: 2006,
            cls.SecurityAccessDenied: 2006,
            cls.AuthenticationRequired: 2020,
            cls.InvalidKey: 2006,
            cls.ExceedNumberOfAttempts: 2006,
            cls.RequiredTimeDelayNotExpired: 2006,
            cls.SecureDataTransmissionRequired: 2020,
            cls.SecureDataTransmissionNotAllowed: 2020,
            cls.SecureDataVerificationFailed: 2020,
            cls.CertificateVerificationFailed_InvalidTimePeriod: 2020,
            cls.CertificateVerificationFailed_InvalidSignature: 2020,
            cls.CertificateVerificationFailed_InvalidChainOfTrust: 2020,
            cls.CertificateVerificationFailed_InvalidType: 2020,
            cls.CertificateVerificationFailed_InvalidFormat: 2020,
            cls.CertificateVerificationFailed_InvalidContent: 2020,
            cls.CertificateVerificationFailed_InvalidScope: 2020,
            cls.CertificateVerificationFailed_InvalidCertificate: 2020,
            cls.OwnershipVerificationFailed: 2020,
            cls.ChallengeCalculationFailed: 2020,
            cls.SettingAccessRightsFailed: 2020,
            cls.SessionKeyCreationDerivationFailed: 2020,
            cls.ConfigurationDataUsageFailed: 2020,
            cls.DeAuthenticationFailed: 2020,
            cls.UploadDownloadNotAccepted: 2006,
            cls.TransferDataSuspended: 2006,
            cls.GeneralProgrammingFailure: 2006,
            cls.WrongBlockSequenceCounter: 2006,
            cls.RequestCorrectlyReceived_ResponsePending: 2006,
            cls.SubFunctionNotSupportedInActiveSession: 2006,
            cls.ServiceNotSupportedInActiveSession: 2006,
            cls.RpmTooHigh: 2006,
            cls.RpmTooLow: 2006,
            cls.EngineIsRunning: 2006,
            cls.EngineIsNotRunning: 2006,
            cls.EngineRunTimeTooLow: 2006,
            cls.TemperatureTooHigh: 2006,
            cls.TemperatureTooLow: 2006,
            cls.VehicleSpeedTooHigh: 2006,
            cls.VehicleSpeedTooLow: 2006,
            cls.ThrottlePedalTooHigh: 2006,
            cls.ThrottlePedalTooLow: 2006,
            cls.TransmissionRangeNotInNeutral: 2006,
            cls.TransmissionRangeNotInGear: 2006,
            cls.BrakeSwitchNotClosed: 2006,
            cls.ShifterLeverNotInPark: 2006,
            cls.TorqueConverterClutchLocked: 2006,
            cls.VoltageTooHigh: 2006,
            cls.VoltageTooLow: 2006,
            cls.ResourceTemporarilyNotAvailable: 2020,
            cls.GeneralSecurityViolation: 2006,
            cls.SecuredModeRequested: 2006,
            cls.InsufficientProtection: 2006,
            cls.TerminationWithSignatureRequested: 2006,
            cls.AccessDenied: 2006,
            cls.VersionNotSupported: 2006,
            cls.SecuredLinkNotSupported: 2006,
            cls.CertificateNotAvailable: 2006,
            cls.AuditTrailInformationNotAvailable: 2006,
        }

        if code not in codes_version:
            raise ValueError('Do not know the standard version in which this code has been introduced: %s' % (code))

        return standard_version >= codes_version[code]

    # Returns the name of the response code as a string

    @classmethod
    def get_name(cls, given_id: int) -> str:
        if given_id is None:
            return ""

        for member in inspect.getmembers(cls):
            if isinstance(member[1], int):
                if member[1] == given_id:
                    return member[0]
        return str(given_id)

    # Tells if a code is a negative code
    @classmethod
    def is_negative(cls, given_id: int) -> bool:
        if given_id in [None, cls.PositiveResponse]:
            return False

        for member in inspect.getmembers(cls):
            if isinstance(member[1], int):
                if member[1] == given_id:
                    return True
        return False
